/***************************************************************************
**                                                                        **
**  Polyphone, a soundfont editor                                         **
**  Copyright (C) 2013-2020 Davy Triponney                                **
**                                                                        **
**  This program is free software: you can redistribute it and/or modify  **
**  it under the terms of the GNU General Public License as published by  **
**  the Free Software Foundation, either version 3 of the License, or     **
**  (at your option) any later version.                                   **
**                                                                        **
**  This program is distributed in the hope that it will be useful,       **
**  but WITHOUT ANY WARRANTY; without even the implied warranty of        **
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the          **
**  GNU General Public License for more details.                          **
**                                                                        **
**  You should have received a copy of the GNU General Public License     **
**  along with this program. If not, see http://www.gnu.org/licenses/.    **
**                                                                        **
****************************************************************************
**           Author: Davy Triponney                                       **
**  Website/Contact: https://www.polyphone-soundfonts.com                 **
**             Date: 01.01.2013                                           **
***************************************************************************/

#include "midi_global_keyboard/extensionmidi_globalkeyboard.h"
#include "midi_global_keyboard/extensionmidi_globalkeyboard_gui.h"
#include "midi_global_keyboard/extensionmidi_globalkeyboard_configuration.h"
#include "midi_global_keyboard/extensionmidi_globalkeyboard_channelselector.h"
#include "contextmanager.h"
#include "synth.h"

ExtensionMidi_GlobalKeyboard::ExtensionMidi_GlobalKeyboard() : ExtensionMidi(),
    _extensionConfig(new ExtensionMidi_GlobalKeyboard_Configuration(getIdentifier())),
    _gui(new ExtensionMidi_GlobalKeyboard_Gui(_extensionConfig))
{
    _triggeredKeys = new quint8 * [16];
    for (int i = 0; i < 16; i++)
    {
        _triggeredKeys[i] = new quint8[128];
        memset(_triggeredKeys[i], 0, 128 * sizeof(quint8));
    }

    connect(_extensionConfig, SIGNAL(muteChannel(quint8)), this, SLOT(onMuteChannel(quint8)));
    connect(_extensionConfig, SIGNAL(presetActivated(quint8, quint8, quint8, quint8, quint8)),
            this, SLOT(onPresetActivated(quint8, quint8, quint8, quint8, quint8)));
    connect(_extensionConfig, SIGNAL(presetDeactivated(quint8, quint8, quint8, quint8, quint8)),
            this, SLOT(onPresetDeactivated(quint8, quint8, quint8, quint8, quint8)));
}

ExtensionMidi_GlobalKeyboard::~ExtensionMidi_GlobalKeyboard()
{
    delete _gui;
    delete _extensionConfig;
    for (int i = 0; i < 16; i++)
        delete [] _triggeredKeys[i];
    delete[] _triggeredKeys;
}

QWidget * ExtensionMidi_GlobalKeyboard::getGui()
{
    return _gui;
}

bool ExtensionMidi_GlobalKeyboard::processKey(int channel, int key, int vel)
{
    if (channel == -1)
        return false;

    if ((vel > 0 && _triggeredKeys[channel][key] > 0) || (vel == 0 && _triggeredKeys[channel][key] == 0))
        return false;
    _triggeredKeys[channel][key] = vel;

    if (vel > 0)
        ExtensionMidi_GlobalKeyboard_ChannelSelector::keyTriggeredForChannel(channel);

    if (!_gui->isVisible() || !_extensionConfig->_usedChannels[channel])
        return false;
    _gui->inputNote(channel, key, vel);

    // Launch or stop voices depending on what is activated for this channel
    processKeyInternal(channel, key, vel);
    return true;
}

void ExtensionMidi_GlobalKeyboard::processKeyInternal(int channel, int key, int vel)
{
    foreach (ActivatedPreset activatedPreset, _extensionConfig->_presetPerChannel[channel])
    {
        if (activatedPreset._state & 0x01) // Normal octave
        {
            if ((activatedPreset._state & 0x02 && key >= 12 && _triggeredKeys[channel][key - 12] > 0) ||
                    (activatedPreset._state & 0x04 && key <= 115 && _triggeredKeys[channel][key + 12] > 0))
            {
                // Do nothing: the sound is played by another key
            }
            else
                ContextManager::audio()->getSynth()->play(EltID(elementPrst, activatedPreset._sf2Id, activatedPreset._presetId), channel, key, vel);
        }
        if ((activatedPreset._state & 0x02) && key <= 115) // Super octave
        {
            if ((activatedPreset._state & 0x01 && _triggeredKeys[channel][key + 12] > 0) ||
                    (activatedPreset._state & 0x04 && key <= 103 && _triggeredKeys[channel][key + 24] > 0))
            {
                // Do nothing: the sound is played by another key
            }
            else
                ContextManager::audio()->getSynth()->play(EltID(elementPrst, activatedPreset._sf2Id, activatedPreset._presetId), channel, key + 12, vel);
        }
        if ((activatedPreset._state & 0x04) && key >= 12) // Sub octave
        {
            if ((activatedPreset._state & 0x01 && _triggeredKeys[channel][key - 12] > 0) ||
                    (activatedPreset._state & 0x02 && key >= 24 && _triggeredKeys[channel][key - 24] > 0))
            {
                // Do nothing: the sound is played by another key
            }
            else
                ContextManager::audio()->getSynth()->play(EltID(elementPrst, activatedPreset._sf2Id, activatedPreset._presetId), channel, key - 12, vel);
        }
    }
}

bool ExtensionMidi_GlobalKeyboard::processPolyPressureChanged(int channel, int key, int pressure)
{
    Q_UNUSED(key)
    Q_UNUSED(pressure)
    if (channel == -1 || !_gui->isVisible())
        return false;

    return _extensionConfig->_usedChannels[channel];
}

bool ExtensionMidi_GlobalKeyboard::processMonoPressureChanged(int channel, int value)
{
    Q_UNUSED(value)
    if (channel == -1 || !_gui->isVisible())
        return false;

    return _extensionConfig->_usedChannels[channel];
}

bool ExtensionMidi_GlobalKeyboard::processControllerChanged(int channel, int num, int value)
{
    Q_UNUSED(num)
    Q_UNUSED(value)
    if (channel == -1 || !_gui->isVisible())
        return false;

    return _extensionConfig->_usedChannels[channel];
}

bool ExtensionMidi_GlobalKeyboard::processBendChanged(int channel, float value)
{
    Q_UNUSED(value)
    if (channel == -1 || !_gui->isVisible())
        return false;

    return _extensionConfig->_usedChannels[channel];
}

bool ExtensionMidi_GlobalKeyboard::processBendSensitivityChanged(int channel, float semitones)
{
    Q_UNUSED(semitones)
    if (channel == -1 || !_gui->isVisible())
        return false;

    return _extensionConfig->_usedChannels[channel];
}

void ExtensionMidi_GlobalKeyboard::onMuteChannel(quint8 channel)
{
    ContextManager::audio()->getSynth()->play(EltID(), channel, -2, 0);
}

void ExtensionMidi_GlobalKeyboard::onPresetActivated(quint8 channel, quint8 sf2Id, quint8 presetId,
                                                     quint8 currentState, quint8 activatedState)
{
    // Load all samples in ram
    SoundfontManager * sm = SoundfontManager::getInstance();
    EltID idPrstInst(elementPrstInst, sf2Id, presetId);
    EltID idInstSmpl(elementInstSmpl, sf2Id);
    EltID idSmpl(elementSmpl, sf2Id);
    foreach (int i, sm->getSiblings(idPrstInst))
    {
        idPrstInst.indexElt2 = i;
        idInstSmpl.indexElt = sm->get(idPrstInst, champ_instrument).wValue;
        foreach (int j, sm->getSiblings(idInstSmpl))
        {
            idInstSmpl.indexElt2 = j;
            idSmpl.indexElt = sm->get(idInstSmpl, champ_sampleID).wValue;
            sm->getSound(idSmpl)->loadInRam();
        }
    }

    // For each key that is already triggered
    for (int key = 0; key < 128; key++)
    {
        if (_triggeredKeys[channel][key] > 0)
        {
            if (activatedState & 0x01) // Normal octave
            {
                if ((currentState & 0x02 && key >= 12 && _triggeredKeys[channel][key - 12] > 0) ||
                        (currentState & 0x04 && key <= 115 && _triggeredKeys[channel][key + 12] > 0))
                {
                    // Do nothing: the sound is played by another key
                }
                else
                    ContextManager::audio()->getSynth()->play(EltID(elementPrst, sf2Id, presetId), channel, key, _triggeredKeys[channel][key]);
            }
            if ((activatedState & 0x02) && key <= 115) // Super octave
            {
                if ((currentState & 0x01 && _triggeredKeys[channel][key + 12] > 0) ||
                        (currentState & 0x04 && key <= 103 && _triggeredKeys[channel][key + 24] > 0))
                {
                    // Do nothing: the sound is played by another key
                }
                else
                    ContextManager::audio()->getSynth()->play(EltID(elementPrst, sf2Id, presetId), channel, key + 12, _triggeredKeys[channel][key]);
            }
            if ((activatedState & 0x04) && key >= 12) // Sub octave
            {
                if ((currentState & 0x01 && _triggeredKeys[channel][key - 12] > 0) ||
                        (currentState & 0x02 && key >= 24 && _triggeredKeys[channel][key - 24] > 0))
                {
                    // Do nothing: the sound is played by another key
                }
                else
                    ContextManager::audio()->getSynth()->play(EltID(elementPrst, sf2Id, presetId), channel, key - 12, _triggeredKeys[channel][key]);
            }
        }
    }
}

void ExtensionMidi_GlobalKeyboard::onPresetDeactivated(quint8 channel, quint8 sf2Id, quint8 presetId,
                                                       quint8 currentState, quint8 deactivatedState)
{
    // For each key that is already triggered
    for (int key = 0; key < 128; key++)
    {
        if (_triggeredKeys[channel][key] > 0)
        {
            if (deactivatedState & 0x01) // Normal octave
            {
                if ((currentState & 0x02 && key >= 12 && _triggeredKeys[channel][key - 12] > 0) ||
                        (currentState & 0x04 && key <= 115 && _triggeredKeys[channel][key + 12] > 0))
                {
                    // Do nothing: the sound is played by another key
                }
                else
                    ContextManager::audio()->getSynth()->play(EltID(elementPrst, sf2Id, presetId), channel, key, 0);
            }
            if ((deactivatedState & 0x02) && key <= 115) // Super octave
            {
                if ((currentState & 0x01 && _triggeredKeys[channel][key + 12] > 0) ||
                        (currentState & 0x04 && key <= 103 && _triggeredKeys[channel][key + 24] > 0))
                {
                    // Do nothing: the sound is played by another key
                }
                else
                    ContextManager::audio()->getSynth()->play(EltID(elementPrst, sf2Id, presetId), channel, key + 12, 0);
            }
            if ((deactivatedState & 0x04) && key >= 12) // Sub octave
            {
                if ((currentState & 0x01 && _triggeredKeys[channel][key - 12] > 0) ||
                        (currentState & 0x02 && key >= 24 && _triggeredKeys[channel][key - 24] > 0))
                {
                    // Do nothing: the sound is played by another key
                }
                else
                    ContextManager::audio()->getSynth()->play(EltID(elementPrst, sf2Id, presetId), channel, key - 12, 0);
            }
        }
    }
}
