/***************************************************************************
**                                                                        **
**  Polyphone, a soundfont editor                                         **
**  Copyright (C) 2013-2020 Davy Triponney                                **
**                                                                        **
**  This program is free software: you can redistribute it and/or modify  **
**  it under the terms of the GNU General Public License as published by  **
**  the Free Software Foundation, either version 3 of the License, or     **
**  (at your option) any later version.                                   **
**                                                                        **
**  This program is distributed in the hope that it will be useful,       **
**  but WITHOUT ANY WARRANTY; without even the implied warranty of        **
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the          **
**  GNU General Public License for more details.                          **
**                                                                        **
**  You should have received a copy of the GNU General Public License     **
**  along with this program. If not, see http://www.gnu.org/licenses/.    **
**                                                                        **
****************************************************************************
**           Author: Davy Triponney                                       **
**  Website/Contact: https://www.polyphone-soundfonts.com                 **
**             Date: 01.01.2013                                           **
***************************************************************************/

#include "extensionmidi_globalkeyboard_channelselector.h"
#include "ui_extensionmidi_globalkeyboard_channelselector.h"
#include "contextmanager.h"

QList<ExtensionMidi_GlobalKeyboard_ChannelSelector *> ExtensionMidi_GlobalKeyboard_ChannelSelector::s_instances =
        QList<ExtensionMidi_GlobalKeyboard_ChannelSelector *>();

ExtensionMidi_GlobalKeyboard_ChannelSelector::ExtensionMidi_GlobalKeyboard_ChannelSelector(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::ExtensionMidi_GlobalKeyboard_ChannelSelector),
    _currentMidiChannel(0)
{
    ui->setupUi(this);

    // Style
    ui->pushScan->setIcon(ContextManager::theme()->getColoredSvg(":/icons/MIDI.svg", QSize(16, 16), ThemeManager::BUTTON_TEXT));
    ui->label->hide();

    // Stop here if there are too many instances
    if (s_instances.count() >= 16)
        return;

    // Populate the channel combobox
    for (int i = 1; i <= 16; i++)
        ui->comboChannel->addItem(tr("Channel %1", "MIDI channel number").arg(i));

    // Find the first channel that is not used
    bool used[16];
    memset(used, 0, 16 * sizeof(bool));
    foreach (ExtensionMidi_GlobalKeyboard_ChannelSelector * instance, s_instances)
        used[instance->_currentMidiChannel] = true;

    for (int i = 0; i < 16; i++)
    {
        if (!used[i])
        {
            _currentMidiChannel = i;
            ui->comboChannel->blockSignals(true);
            ui->comboChannel->setCurrentIndex(i);
            ui->comboChannel->blockSignals(false);
            break;
        }
    }

    s_instances << this;

    connect(ui->pushScan, SIGNAL(focusOut()), this, SLOT(onFocusOut()));
}

ExtensionMidi_GlobalKeyboard_ChannelSelector::~ExtensionMidi_GlobalKeyboard_ChannelSelector()
{
    s_instances.removeOne(this);
    delete ui;
}

void ExtensionMidi_GlobalKeyboard_ChannelSelector::on_pushScan_clicked()
{
    changeState(ui->comboChannel->isVisible());
}

void ExtensionMidi_GlobalKeyboard_ChannelSelector::on_comboChannel_currentIndexChanged(int index)
{
    // Check the channel is not already selected in other instance
    foreach (ExtensionMidi_GlobalKeyboard_ChannelSelector * instance, s_instances)
    {
        if (instance->getMidiChannel() == index)
        {
            // Give the current channel to the other instance
            instance->ui->comboChannel->blockSignals(true);
            instance->ui->comboChannel->setCurrentIndex(_currentMidiChannel);
            instance->ui->comboChannel->blockSignals(false);
            instance->_currentMidiChannel = _currentMidiChannel;
            emit(instance->midiChannelChanged(_currentMidiChannel, true));
            break;
        }
    }
    _currentMidiChannel = index;
    emit(midiChannelChanged(_currentMidiChannel, false));
}

void ExtensionMidi_GlobalKeyboard_ChannelSelector::onFocusOut()
{
    changeState(false);
}

void ExtensionMidi_GlobalKeyboard_ChannelSelector::changeState(bool isScanMode)
{
    if (isScanMode)
    {
        ui->comboChannel->hide();
        ui->label->show();
    }
    else
    {
        ui->label->hide();
        ui->comboChannel->show();
    }
}

QSize ExtensionMidi_GlobalKeyboard_ChannelSelector::minimumSizeHint() const
{
    QSize size = QWidget::minimumSizeHint();
    size.setHeight(ui->comboChannel->height());
    return size;
}

void ExtensionMidi_GlobalKeyboard_ChannelSelector::setMidiChannel(int channel)
{
    ui->comboChannel->blockSignals(true);
    ui->comboChannel->setCurrentIndex(channel);
    ui->comboChannel->blockSignals(false);
    _currentMidiChannel = channel;
}

void ExtensionMidi_GlobalKeyboard_ChannelSelector::keyTriggeredForChannel(int channel)
{
    if (channel == -1)
        return;

    foreach (ExtensionMidi_GlobalKeyboard_ChannelSelector * instance, s_instances)
    {
        if (instance->ui->label->isVisible())
        {
            // Change the current channel for this instance
            instance->changeState(false);
            instance->ui->comboChannel->setCurrentIndex(channel);
            break;
        }
    }
}
