/***************************************************************************
**                                                                        **
**  Polyphone, a soundfont editor                                         **
**  Copyright (C) 2013-2020 Davy Triponney                                **
**                                                                        **
**  This program is free software: you can redistribute it and/or modify  **
**  it under the terms of the GNU General Public License as published by  **
**  the Free Software Foundation, either version 3 of the License, or     **
**  (at your option) any later version.                                   **
**                                                                        **
**  This program is distributed in the hope that it will be useful,       **
**  but WITHOUT ANY WARRANTY; without even the implied warranty of        **
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the          **
**  GNU General Public License for more details.                          **
**                                                                        **
**  You should have received a copy of the GNU General Public License     **
**  along with this program. If not, see http://www.gnu.org/licenses/.    **
**                                                                        **
****************************************************************************
**           Author: Davy Triponney                                       **
**  Website/Contact: https://www.polyphone-soundfonts.com                 **
**             Date: 01.01.2013                                           **
***************************************************************************/

#include "extensionmidi_globalkeyboard_configuration.h"
#include "contextmanager.h"

ExtensionMidi_GlobalKeyboard_Configuration::ExtensionMidi_GlobalKeyboard_Configuration(QString extensionIdentifier) : QObject(),
    _extensionIdentifier(extensionIdentifier)
{
    // Default: no keyboards and all channels are unused
    for (int i = 0; i < 16; i++)
        _channelByKeyboard[i] = -1;
    memset(_usedChannels, 0, 16 * sizeof(bool));

    // Load the configuration if something is stored
    loadChannelsFromConfig();
}

ExtensionMidi_GlobalKeyboard_Configuration::~ExtensionMidi_GlobalKeyboard_Configuration()
{
}

void ExtensionMidi_GlobalKeyboard_Configuration::addChannel(int channel)
{
    // Update the channel per keyboard
    for (int i = 0; i < 16; i++)
    {
        if (_channelByKeyboard[i] == -1)
        {
            _channelByKeyboard[i] = channel;
            break;
        }
    }

    // Update the used channels
    _usedChannels[channel] = true;

    // Store the channel per keyboard configuration
    saveChannelsToConfig();
}

void ExtensionMidi_GlobalKeyboard_Configuration::removeChannel(int channel)
{
    emit(muteChannel(channel));

    // Update the used channels
    _usedChannels[channel] = false;

    // Update the channel per keyboard
    for (int i = 0; i < 16; i++)
    {
        if (_channelByKeyboard[i] == channel)
        {
            for (int j = i; j < 16 - 1; j++)
                _channelByKeyboard[j] = _channelByKeyboard[j + 1];
            _channelByKeyboard[15] = -1;
            break;
        }
    }

    // Store the channel per keyboard configuration
    saveChannelsToConfig();
}

void ExtensionMidi_GlobalKeyboard_Configuration::setChannels(QList<int> channels)
{
    // Update the channel per keyboard
    int swappedConf = -1;
    for (int i = 0; i < channels.count(); i++)
    {
        if (_channelByKeyboard[i] != channels[i])
        {
            // Swap the activated presets
            if (_channelByKeyboard[i] != swappedConf)
            {
                QVector<ActivatedPreset> tmp = _presetPerChannel[channels[i]];
                _presetPerChannel[channels[i]] = _presetPerChannel[_channelByKeyboard[i]];
                _presetPerChannel[_channelByKeyboard[i]] = tmp;
                swappedConf = channels[i];
            }

            _channelByKeyboard[i] = channels[i];
        }
    }

    // Update the used channels
    for (int i = 0; i < 16; i++)
        _usedChannels[i] = channels.contains(i);

    // Store the channel per keyboard configuration
    saveChannelsToConfig();
}

void ExtensionMidi_GlobalKeyboard_Configuration::loadChannelsFromConfig()
{
    QList<QVariant> list = ContextManager::configuration()->getExtensionValue(_extensionIdentifier, "channels", QVariantList()).toList();
    for (int i = 0; i < list.count(); i++)
    {
        _channelByKeyboard[i] = list[i].toInt();
        _usedChannels[list[i].toInt()] = true;
    }
}

void ExtensionMidi_GlobalKeyboard_Configuration::saveChannelsToConfig()
{
    QList<QVariant> listRet;
    for (int i = 0; i < 16; i++)
    {
        if (_channelByKeyboard[i] == -1)
            break;
        listRet << _channelByKeyboard[i];
    }
    ContextManager::configuration()->setExtensionValue(_extensionIdentifier, "channels", listRet);
}

quint8 ExtensionMidi_GlobalKeyboard_Configuration::getActivationState(quint8 channel, quint8 sf2Id, quint8 presetId)
{
    foreach (ActivatedPreset activatedPreset, _presetPerChannel[channel])
    {
        if (activatedPreset._sf2Id == sf2Id && activatedPreset._presetId == presetId)
            return activatedPreset._state;
    }
    return 0;
}

void ExtensionMidi_GlobalKeyboard_Configuration::setActivationState(quint8 channel, quint8 sf2Id, quint8 presetId, quint8 state)
{
    for (int i = 0; i < _presetPerChannel[channel].count(); i++)
    {
        if (_presetPerChannel[channel][i]._sf2Id == sf2Id && _presetPerChannel[channel][i]._presetId == presetId)
        {
            if (state == 0)
            {
                state = _presetPerChannel[channel][i]._state;
                _presetPerChannel[channel].removeAt(i);
                emit(presetDeactivated(channel, sf2Id, presetId, state, state));
            }
            else
            {
                quint8 initialState = _presetPerChannel[channel][i]._state;
                _presetPerChannel[channel][i]._state = state;
                if (state > initialState)
                    emit(presetActivated(channel, sf2Id, presetId, initialState, state - initialState));
                else
                    emit(presetDeactivated(channel, sf2Id, presetId, initialState, initialState - state));
            }
            return;
        }
    }

    if (state != 0)
    {
        _presetPerChannel[channel] << ActivatedPreset(sf2Id, presetId, state);
        emit(presetActivated(channel, sf2Id, presetId, 0, state));
    }
}

void ExtensionMidi_GlobalKeyboard_Configuration::onSoundfontClosed(int sf2Index)
{
    // Remove all elements pointing to a specific sf2Index
    for (int i = 0; i < 16; i++)
    {
        for (int j = _presetPerChannel[i].count() - 1; j >= 0; j--)
        {
            if (_presetPerChannel[i][j]._sf2Id == sf2Index)
                _presetPerChannel[i].removeAt(j);
        }
    }
}
