/***************************************************************************
**                                                                        **
**  Polyphone, a soundfont editor                                         **
**  Copyright (C) 2013-2020 Davy Triponney                                **
**                                                                        **
**  This program is free software: you can redistribute it and/or modify  **
**  it under the terms of the GNU General Public License as published by  **
**  the Free Software Foundation, either version 3 of the License, or     **
**  (at your option) any later version.                                   **
**                                                                        **
**  This program is distributed in the hope that it will be useful,       **
**  but WITHOUT ANY WARRANTY; without even the implied warranty of        **
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the          **
**  GNU General Public License for more details.                          **
**                                                                        **
**  You should have received a copy of the GNU General Public License     **
**  along with this program. If not, see http://www.gnu.org/licenses/.    **
**                                                                        **
****************************************************************************
**           Author: Davy Triponney                                       **
**  Website/Contact: https://www.polyphone-soundfonts.com                 **
**             Date: 01.01.2013                                           **
***************************************************************************/

#include "extensionmidi_globalkeyboard_gui.h"
#include "ui_extensionmidi_globalkeyboard_gui.h"
#include "midi_global_keyboard/extensionmidi_globalkeyboard_keyboardcell.h"
#include "midi_global_keyboard/extensionmidi_globalkeyboard_presettable.h"
#include "midi_global_keyboard/extensionmidi_globalkeyboard_splitter.h"
#include "midi_global_keyboard/extensionmidi_globalkeyboard_configuration.h"
#include "contextmanager.h"
#include <QPushButton>
#include <QLabel>

ExtensionMidi_GlobalKeyboard_Gui::ExtensionMidi_GlobalKeyboard_Gui(ExtensionMidi_GlobalKeyboard_Configuration * extensionConfig) : QWidget(),
    ui(new Ui::ExtensionMidi_GlobalKeyboard_Gui),
    _extensionConfig(extensionConfig)
{
    ui->setupUi(this);

    // QSplitter for being able to resize the tree
    ExtensionMidi_GlobalKeyboard_Splitter * splitter = new ExtensionMidi_GlobalKeyboard_Splitter(
                this, ui->listKeyboards, ui->tablePresets, _extensionConfig->_extensionIdentifier);
    QHBoxLayout * layout = dynamic_cast<QHBoxLayout *>(this->layout());
    layout->addWidget(splitter);
    connect(splitter, SIGNAL(splitterMoved(int,int)), this, SLOT(onSplitterMoved(int,int)));

    // Style
    ui->listKeyboards->setStyleSheet("QListWidget{background:" +
                                     ContextManager::theme()->getColor(ThemeManager::WINDOW_BACKGROUND).name() +
                                     ";show-decoration-selected: 0;}");

    // Refresh the table everytime a soundfont is edited
    connect(SoundfontManager::getInstance(), SIGNAL(editingDone(QString,QList<int>)), this, SLOT(onEditingDone(QString,QList<int>)));
    connect(SoundfontManager::getInstance(), SIGNAL(soundfontLoaded(int)), this, SLOT(onSoundfontLoaded(int)));
    connect(SoundfontManager::getInstance(), SIGNAL(soundfontClosed(int)), this, SLOT(onSoundfontClosed(int)));

    // Add a button "+"
    _newKeyboardButton = new QPushButton(ui->listKeyboards);
    _newKeyboardButton->setIcon(ContextManager::theme()->getColoredSvg(":/icons/document-new.svg", QSize(16, 16), ThemeManager::BUTTON_TEXT));
    _newKeyboardButton->setToolTip(tr("Add keyboard"));
    connect(_newKeyboardButton, SIGNAL(clicked(bool)), this, SLOT(onAddKeyboardClicked(bool)));
    QHBoxLayout * hLayout = new QHBoxLayout();
    hLayout->addWidget(_newKeyboardButton);
    hLayout->addStretch();
    QWidget * w = new QWidget();
    w->setLayout(hLayout);

    QListWidgetItem * item = new QListWidgetItem();
    item->setSizeHint(w->sizeHint());
    ui->listKeyboards->addItem(item);
    ui->listKeyboards->setItemWidget(item, w);

    // Load the keyboards with the corresponding channels
    bool withKeyboard = false;
    for (int i = 0; i < 16; i++)
    {
        if (_extensionConfig->_channelByKeyboard[i] == -1)
            break;
        addKeyboard(_extensionConfig->_channelByKeyboard[i]);
        withKeyboard = true;
    }

    // Add at least one keyboard
    if (withKeyboard)
    {
        // Update the keyboard list
        updateCellHeight();

        // Update the preset table
        ui->tablePresets->updateTable(_extensionConfig);
    }
    else
        addKeyboard();
}

ExtensionMidi_GlobalKeyboard_Gui::~ExtensionMidi_GlobalKeyboard_Gui()
{
    delete ui;
}

void ExtensionMidi_GlobalKeyboard_Gui::onSplitterMoved(int pos, int index)
{
    Q_UNUSED(pos)
    Q_UNUSED(index)
    updateCellHeight();
}

void ExtensionMidi_GlobalKeyboard_Gui::resizeEvent(QResizeEvent * event)
{
    QWidget::resizeEvent(event);
    updateCellHeight();
}

void ExtensionMidi_GlobalKeyboard_Gui::updateCellHeight()
{
    // Update size hints for all cells
    int viewPortWidth = ui->listKeyboards->viewport()->width();
    for (int i = 0; i < ui->listKeyboards->count() - 1; i++)
    {
        QListWidgetItem * item = ui->listKeyboards->item(i);
        ExtensionMidi_GlobalKeyboard_KeyboardCell* cell = dynamic_cast<ExtensionMidi_GlobalKeyboard_KeyboardCell*>(ui->listKeyboards->itemWidget(item));
        cell->setMaximumWidth(viewPortWidth);
        item->setSizeHint(QSize(viewPortWidth, cell->heightForWidth(viewPortWidth)));
    }
}

void ExtensionMidi_GlobalKeyboard_Gui::onEditingDone(QString editingSource, QList<int> sf2Indexes)
{
    Q_UNUSED(editingSource)
    Q_UNUSED(sf2Indexes)

    // Update the preset table
    ui->tablePresets->updateTable(_extensionConfig);
}

void ExtensionMidi_GlobalKeyboard_Gui::onSoundfontLoaded(int sf2Index)
{
    Q_UNUSED(sf2Index)

    // Update the preset table
    ui->tablePresets->updateTable(_extensionConfig);
}

void ExtensionMidi_GlobalKeyboard_Gui::onSoundfontClosed(int sf2Index)
{
    _extensionConfig->onSoundfontClosed(sf2Index);

    // Update the preset table
    ui->tablePresets->updateTable(_extensionConfig);
}

void ExtensionMidi_GlobalKeyboard_Gui::onAddKeyboardClicked(bool checked)
{
    Q_UNUSED(checked)
    addKeyboard();
}

void ExtensionMidi_GlobalKeyboard_Gui::onChannelChanged()
{
    // Get the current channel list
    QList<int> channels;
    for (int i = 0; i < ui->listKeyboards->count() - 1; i++)
    {
        QListWidgetItem * item = ui->listKeyboards->item(i);
        ExtensionMidi_GlobalKeyboard_KeyboardCell* cell = dynamic_cast<ExtensionMidi_GlobalKeyboard_KeyboardCell*>(ui->listKeyboards->itemWidget(item));
        channels << cell->getChannel();
    }

    // Update the configuration
    _extensionConfig->setChannels(channels);

    // Update the preset table
    ui->tablePresets->updateTable(_extensionConfig);
}

void ExtensionMidi_GlobalKeyboard_Gui::onPushRemoveClicked()
{
    ExtensionMidi_GlobalKeyboard_KeyboardCell * sender = reinterpret_cast<ExtensionMidi_GlobalKeyboard_KeyboardCell*>(QObject::sender());

    // Remove a keyboard
    for (int i = 0; i < ui->listKeyboards->count() - 1; i++)
    {
        QListWidgetItem * item = ui->listKeyboards->item(i);
        ExtensionMidi_GlobalKeyboard_KeyboardCell* cell = dynamic_cast<ExtensionMidi_GlobalKeyboard_KeyboardCell*>(ui->listKeyboards->itemWidget(item));
        if (cell == sender)
        {
            delete ui->listKeyboards->takeItem(i);
            break;
        }
    }

    // Update the configuration
    _extensionConfig->removeChannel(sender->getChannel());

    // Update the preset table
    ui->tablePresets->updateTable(_extensionConfig);
}

void ExtensionMidi_GlobalKeyboard_Gui::addKeyboard(int channel)
{
    ExtensionMidi_GlobalKeyboard_KeyboardCell * cell = new ExtensionMidi_GlobalKeyboard_KeyboardCell(channel, ui->listKeyboards);
    connect(cell, SIGNAL(channelChanged()), this, SLOT(onChannelChanged()));
    connect(cell, SIGNAL(pushRemoveClicked()), this, SLOT(onPushRemoveClicked()));
    if (ui->listKeyboards->count() == 1)
        cell->disableRemoveButton();

    QListWidgetItem * item = new QListWidgetItem();
    ui->listKeyboards->insertItem(ui->listKeyboards->count() - 1, item);
    ui->listKeyboards->setItemWidget(item, cell);

    if (ui->listKeyboards->count() >= 17)
        _newKeyboardButton->setEnabled(false);

    if (channel == -1)
    {
        // Update the configuration
        _extensionConfig->addChannel(cell->getChannel());

        // Update the keyboard list
        updateCellHeight();

        // Update the preset table
        ui->tablePresets->updateTable(_extensionConfig);
    }
}

void ExtensionMidi_GlobalKeyboard_Gui::inputNote(int channel, int key, int vel)
{
    for (int i = 0; i < ui->listKeyboards->count() - 1; i++)
    {
        QListWidgetItem * item = ui->listKeyboards->item(i);
        ExtensionMidi_GlobalKeyboard_KeyboardCell * cell = dynamic_cast<ExtensionMidi_GlobalKeyboard_KeyboardCell *>(ui->listKeyboards->itemWidget(item));
        if (cell->getChannel() == channel)
        {
            cell->inputNote(key, vel);
            break;
        }
    }
}
