/***************************************************************************
**                                                                        **
**  Polyphone, a soundfont editor                                         **
**  Copyright (C) 2013-2020 Davy Triponney                                **
**                                                                        **
**  This program is free software: you can redistribute it and/or modify  **
**  it under the terms of the GNU General Public License as published by  **
**  the Free Software Foundation, either version 3 of the License, or     **
**  (at your option) any later version.                                   **
**                                                                        **
**  This program is distributed in the hope that it will be useful,       **
**  but WITHOUT ANY WARRANTY; without even the implied warranty of        **
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the          **
**  GNU General Public License for more details.                          **
**                                                                        **
**  You should have received a copy of the GNU General Public License     **
**  along with this program. If not, see http://www.gnu.org/licenses/.    **
**                                                                        **
****************************************************************************
**           Author: Davy Triponney                                       **
**  Website/Contact: https://www.polyphone-soundfonts.com                 **
**             Date: 01.01.2013                                           **
***************************************************************************/

#include "extensionmidi_globalkeyboard_presettable.h"
#include "midi_global_keyboard/extensionmidi_globalkeyboard_presetcell.h"
#include "midi_global_keyboard/extensionmidi_globalkeyboard_configuration.h"
#include "soundfontmanager.h"
#include "contextmanager.h"
#include <QHeaderView>
#include <QPushButton>

ExtensionMidi_GlobalKeyboard_PresetTable::ExtensionMidi_GlobalKeyboard_PresetTable(QWidget * parent) : QTableWidget(parent)
{
    this->verticalHeader()->hide();
    this->horizontalHeader()->setSectionResizeMode(QHeaderView::Fixed);
    _clearColor = ContextManager::theme()->mix(ContextManager::theme()->getColor(ThemeManager::LIST_BACKGROUND),
                                               ContextManager::theme()->getColor(ThemeManager::LIST_TEXT), 0.5).name();
}

void ExtensionMidi_GlobalKeyboard_PresetTable::updateTable(ExtensionMidi_GlobalKeyboard_Configuration * extensionConfig)
{
    QFont fontBold = this->font();
    fontBold.setBold(true);
    fontBold.setPointSize(fontBold.pointSize() + 10);
    SoundfontManager * sm = SoundfontManager::getInstance();
    QTableWidgetItem * item;

    // Clear everything
    _cellBySf2AndChannel.clear();
    this->clear();
    this->setRowCount(0);

    // Number of colums, horizontal header labels
    int channelCount = 16;
    QStringList headerLabels = QStringList() << "" << "";
    for (int i = 0; i < 16; i++)
    {
        if (extensionConfig->_channelByKeyboard[i] == -1)
        {
            channelCount = i;
            break;
        }
        else
            headerLabels << "Channel " + QString::number(extensionConfig->_channelByKeyboard[i] + 1);
    }

    this->setColumnCount(2 + channelCount);
    this->setHorizontalHeaderLabels(headerLabels);

    // For each soundfont
    EltID idSf2(elementSf2);
    foreach (int indexSf2, sm->getSiblings(idSf2))
    {
        // Name of the soundfont
        idSf2.indexSf2 = indexSf2;
        QString name = sm->getQstr(idSf2, champ_name);
        if (name.isEmpty())
            name = sm->getQstr(idSf2, champ_filenameInitial).split(QRegularExpression("(/|\\\\)")).last();
        if (name.isEmpty())
            name = tr("Untitled");

        // Add a title in the table
        this->insertRow(this->rowCount());
        this->setSpan(this->rowCount() - 1, 0, 1, 2);
        item = new QTableWidgetItem(name);
        item->setFont(fontBold);
        this->setItem(this->rowCount() - 1, 0, item);
        this->setRowHeight(this->rowCount() - 1, QFontMetrics(fontBold).height() + 12);

        // Add the "clear" button
        for (int i = 0; i < channelCount; i++)
        {
            QPushButton * button = new QPushButton();
            button->setText(tr("Clear"));
            button->setFlat(true);
            button->setCursor(Qt::PointingHandCursor);
            button->setStyleSheet("QPushButton{color:" + _clearColor + ";}");
            connect(button, SIGNAL(clicked(bool)), this, SLOT(onClearClicked(bool)));

            button->setProperty("channel", i);
            button->setProperty("sf2Index", indexSf2);

            this->setCellWidget(this->rowCount() - 1, 2 + i, button);
        }

        // Preset list
        QList<PresetRow> presetRows = getPresetList(indexSf2);
        foreach (PresetRow presetRow, presetRows)
        {
            this->insertRow(this->rowCount());

            // Column 1: bank / preset
            QString presetNumber = QString("%1:%2")
                    .arg(presetRow._bank, 3, 10, QChar('0'))
                    .arg(presetRow._preset, 3, 10, QChar('0'));
            this->setItem(this->rowCount() - 1, 0, new QTableWidgetItem(presetNumber));

            // Column 2: name
            item = new QTableWidgetItem(presetRow._name);
            this->setItem(this->rowCount() - 1, 1, item);

            // Following columns: checkboxes for enabling the preset to be played with a specific channel
            for (int i = 0; i < channelCount; i++)
            {
                ExtensionMidi_GlobalKeyboard_PresetCell * cell =
                        new ExtensionMidi_GlobalKeyboard_PresetCell(
                            extensionConfig, extensionConfig->_channelByKeyboard[i],
                            indexSf2, presetRow._indexElt);
                _cellBySf2AndChannel[indexSf2][i] << cell;
                this->setCellWidget(this->rowCount() - 1, 2 + i, cell);
            }
        }
    }

    this->resizeColumnsToContents();
}

QList<ExtensionMidi_GlobalKeyboard_PresetTable::PresetRow> ExtensionMidi_GlobalKeyboard_PresetTable::getPresetList(int indexSf2)
{
    SoundfontManager * sm = SoundfontManager::getInstance();
    QList<PresetRow> list;

    // Browse all presets
    EltID idPrst(elementPrst, indexSf2);
    foreach (int prstId, sm->getSiblings(idPrst))
    {
        idPrst.indexElt = prstId;
        list << PresetRow(prstId,
                          sm->get(idPrst, champ_wBank).wValue,
                          sm->get(idPrst, champ_wPreset).wValue,
                          sm->getQstr(idPrst, champ_name));
    }

    // Sort the list
    std::sort(list.begin(), list.end(),
              [](const PresetRow& elt1 , const PresetRow& elt2)->bool
    {
        if (elt1._bank == elt2._bank)
            return elt1._preset < elt2._preset;
        return elt1._bank < elt2._bank;
    });

    return list;
}

void ExtensionMidi_GlobalKeyboard_PresetTable::onClearClicked(bool checked)
{
    Q_UNUSED(checked)

    // Channel and sf2 index to clear
    QPushButton * button = reinterpret_cast<QPushButton *>(QObject::sender());
    int channel = button->property("channel").toInt();
    int sf2Index = button->property("sf2Index").toInt();

    // Remove all references to sf2Index by channel
    foreach (ExtensionMidi_GlobalKeyboard_PresetCell * cell, _cellBySf2AndChannel[sf2Index][channel])
        cell->clear();
}
