/***************************************************************************
**                                                                        **
**  Polyphone, a soundfont editor                                         **
**  Copyright (C) 2013-2020 Davy Triponney                                **
**                                                                        **
**  This program is free software: you can redistribute it and/or modify  **
**  it under the terms of the GNU General Public License as published by  **
**  the Free Software Foundation, either version 3 of the License, or     **
**  (at your option) any later version.                                   **
**                                                                        **
**  This program is distributed in the hope that it will be useful,       **
**  but WITHOUT ANY WARRANTY; without even the implied warranty of        **
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the          **
**  GNU General Public License for more details.                          **
**                                                                        **
**  You should have received a copy of the GNU General Public License     **
**  along with this program. If not, see http://www.gnu.org/licenses/.    **
**                                                                        **
****************************************************************************
**           Author: Davy Triponney                                       **
**  Website/Contact: https://www.polyphone-soundfonts.com                 **
**             Date: 01.01.2013                                           **
***************************************************************************/

#include "extensionview_bars_graphics.h"
#include "soundfontmanager.h"
#include "contextmanager.h"
#include "extensionview_bars_graphics_rectangleitem.h"
#include "extensionview_bars_graphics_legenditem.h"
#include "extensionview_bars_graphics_legenditem2.h"
#include "extensionview_bars_graphics_key.h"
#include "graphicssimpletextitem.h"
#include "noteevent.h"
#include <QMouseEvent>
#include <QApplication>

// Z values (all GraphicsItem being at the top level)
//   0: grid
//  25: play marker
//  50: non selected rectangles
//  51: selected rectangles
// 100: axis values
// 120: legends

const int ExtensionView_Bars_Graphics::HLINE_NUMBER = 11;

ExtensionView_Bars_Graphics::ExtensionView_Bars_Graphics(QWidget * parent) : QGraphicsView(parent),
    _sf2(SoundfontManager::getInstance()),
    _scene(new QGraphicsScene()),
    _legendItem(nullptr),
    _legendItem2(nullptr),
    _keyTriggered(-1),
    _removeIndexOnRelease(false),
    _removeOtherIndexOnRelease(false),
    _currentMode(NORMAL_MODE),
    _keepFirstShiftPoint(false),
    _yInit(0),
    _minKey(36),
    _maxKey(96),
    _firstShiftRect(-100, 0, 0, 0)
{
    // Colors
    QColor color = ContextManager::theme()->getColor(ThemeManager::LIST_TEXT);
    color.setAlpha(180);
    _textColor = color;
    color.setAlpha(40);
    _lineColor = color;

    // Configuration
    this->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    this->setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    this->setRenderHint(QPainter::Antialiasing, true);
    this->setMouseTracking(true);

    // Preparation of the graphics
    this->setScene(_scene);
    this->initItems();
}

ExtensionView_Bars_Graphics::~ExtensionView_Bars_Graphics()
{
    while (!_rectangles.isEmpty())
        delete _rectangles.takeFirst();
    while (!_leftLabels.isEmpty())
        delete _leftLabels.takeFirst();
    while (!_bottomLabels.isEmpty())
        delete _bottomLabels.takeFirst();
    delete _legendItem;
    delete _legendItem2;
    while (!_vLines.isEmpty())
        delete _vLines.takeFirst();
    while (!_hLines.isEmpty())
        delete _hLines.takeFirst();
    while (!_mapGraphicsKeys.isEmpty())
        delete _mapGraphicsKeys.take(_mapGraphicsKeys.keys().first());
    delete _scene;
}

void ExtensionView_Bars_Graphics::initItems()
{
    // Vertical lines
    QPen penVerticalLines(_lineColor, 1);
    penVerticalLines.setCosmetic(true);
    for (quint32 key = 12; key <= 120; key += 12)
    {
        QGraphicsLineItem * line = new QGraphicsLineItem(key, -2.0, key, 2.0);
        _scene->addItem(line);
        line->setPen(penVerticalLines);
        line->setZValue(0);
        _vLines << line;

        GraphicsSimpleTextItem * text = new GraphicsSimpleTextItem(Qt::AlignHCenter | Qt::AlignBottom);
        _scene->addItem(text);
        text->setZValue(100);
        text->setBrush(_textColor);
        text->setText(ContextManager::keyName()->getKeyName(key));
        text->setPos(key, 1.02);
        _bottomLabels << text;
    }

    // Horizontal lines: prepare lines and labels
    QPen penHorizontalLines(_lineColor, 1, Qt::DotLine);
    penHorizontalLines.setCosmetic(true);
    for (int i = 0; i < HLINE_NUMBER; i++)
    {
        QGraphicsLineItem * line = new QGraphicsLineItem(-1, 0, 128, 0);
        _scene->addItem(line);
        line->setPen(penHorizontalLines);
        line->setZValue(0);
        _hLines << line;

        GraphicsSimpleTextItem * text = new GraphicsSimpleTextItem(Qt::AlignLeft | Qt::AlignVCenter);
        _scene->addItem(text);
        text->setZValue(100);
        text->setBrush(_textColor);
        _leftLabels << text;
    }

    // Legends
    _legendItem = new ExtensionView_Bars_Graphics_LegendItem(this->font().family());
    _scene->addItem(_legendItem);
    _legendItem->setZValue(120);
    _legendItem2 = new ExtensionView_Bars_Graphics_LegendItem2(this->font().family());
    _scene->addItem(_legendItem2);
    _legendItem2->setZValue(120);
}

void ExtensionView_Bars_Graphics::updateLabelPosition()
{
    // Update the position of the axis labels (they stay to the left and bottom)
    foreach (GraphicsSimpleTextItem * label, _leftLabels)
        label->setX(_minKey - 0.5);

    // Update the position of the legend (it stays in a corner)
    if (_legendItem->isLeft())
    {
        _legendItem->setX(_minKey);
        _legendItem2->setX(_minKey);
    }
    else
    {
        _legendItem->setX(_maxKey);
        _legendItem2->setX(_maxKey);
    }
}

void ExtensionView_Bars_Graphics::updateHover(QPoint mousePos)
{
    if (_currentMode == EDITING_CONTOUR_MODE)
    {
        _legendItem->setIds(QList<EltID>(), QList<int>(), 0, 0);
        return;
    }

    // Rectangles under the mouse
    QList<QList<ExtensionView_Bars_Graphics_RectangleItem *> > pairs = getRectanglesUnderMouse(mousePos);

    // Highlighted rectangles under the mouse
    QList<ExtensionView_Bars_Graphics_RectangleItem*> hoveredRectangles;
    int focussedIndex = 0;
    if (pairs.count() > 0)
    {
        int focussedAndSelected = -1;
        int focussed = -1;
        int selected = -1;
        for (int i = 0; i < pairs.count(); i++)
        {
            foreach (ExtensionView_Bars_Graphics_RectangleItem * item, pairs[i])
            {
                if (item->isHovered())
                {
                    if (item->isSelected())
                        focussedAndSelected = i;
                    else
                        focussed = i;
                }
                else
                {
                    if (item->isSelected())
                        selected = i;
                }
            }
        }

        // Priority for the focussed rectangles, then selected
        focussedIndex = (focussedAndSelected != -1 ? focussedAndSelected : (focussed != -1 ? focussed : (selected != -1 ? selected : 0)));
        hoveredRectangles = pairs[focussedIndex];
    }

    // Update the hover, get the type of editing in the same time
    foreach (ExtensionView_Bars_Graphics_RectangleItem * item, _rectangles)
        item->setHover(hoveredRectangles.contains(item));

    // Update the content of the first legend
    QList<EltID> ids;
    QList<int> selectedIds;
    int count = 0;
    for (int i = 0; i < pairs.count(); i++)
    {
        foreach (ExtensionView_Bars_Graphics_RectangleItem * item, pairs[i])
        {
            ids << item->getID();
            if (i == focussedIndex)
                selectedIds << count;
            count++;
        }
    }

    _legendItem->setIds(ids, selectedIds, focussedIndex, pairs.count());

    // Offset and location of the second legend
    _legendItem2->setOffsetY(_legendItem->boundingRect().bottom());
}

void ExtensionView_Bars_Graphics::display(IdList ids, AttributeType attributeType, bool justSelection)
{
    if (!justSelection)
    {
        // Clear previous rectangles
        while (!_rectangles.isEmpty())
        {
            _scene->removeItem(_rectangles.first());
            delete _rectangles.takeFirst();
        }

        // Add new ones, compute key min and key max
        _defaultID = ids[0];
        EltID idDiv = ids[0];
        bool isPrst = true;
        switch (_defaultID.typeElement)
        {
        case elementInst: case elementInstSmpl:
            _defaultID.typeElement = elementInst;
            idDiv.typeElement = elementInstSmpl;
            isPrst = false;
            break;
        case elementPrst: case elementPrstInst:
            _defaultID.typeElement = elementPrst;
            idDiv.typeElement = elementPrstInst;
            break;
        default:
            return;
        }

        _minKey = 36;
        _maxKey = 96;
        foreach (int i, _sf2->getSiblings(idDiv))
        {
            idDiv.indexElt2 = i;
            RangesType keyRange = _sf2->get(idDiv, champ_keyRange).rValue;
            if (keyRange.byLo < _minKey)
                _minKey = keyRange.byLo;
            if (keyRange.byHi > _maxKey)
                _maxKey = keyRange.byHi;

            ExtensionView_Bars_Graphics_RectangleItem * item = new ExtensionView_Bars_Graphics_RectangleItem(idDiv, attributeType);
            item->setZValue(50);
            _scene->addItem(item);
            _rectangles << item;
        }

        // Scene rect
        _displayedRect = QRectF(_minKey - 0.5, -0.01, (_maxKey - _minKey + 1), 1.03);
        _scene->setSceneRect(_displayedRect);
        fitInView(_displayedRect);

        // Legend configuration
        _legendItem->setAttribute(attributeType, isPrst);
        _legendItem2->setAttribute(attributeType, isPrst);

        // Vertical lines and labels
        updateHorizontalLines(attributeType, isPrst);

        updateLabelPosition();
    }

    // Selection
    foreach (ExtensionView_Bars_Graphics_RectangleItem * item, _rectangles)
    {
        if (ids.contains(item->getID()))
        {
            item->setSelected(true);
            if (!_keepFirstShiftPoint)
                _firstShiftRect = item->rect();
        }
        else
            item->setSelected(false);
    }

    viewport()->update();
}

void ExtensionView_Bars_Graphics::updateHorizontalLines(AttributeType attributeType, bool isPrst)
{
    // First hide all lines
    for (int i = 0; i < HLINE_NUMBER; i++)
    {
        _leftLabels[i]->setY(-1);
        _hLines[i]->setY(-1);
    }

    // Get the left values and the unit
    QList<double> leftValues;
    QString unit = "";
    switch (attributeType)
    {
    case champ_initialAttenuation:
        if (isPrst)
            leftValues << -50.0 << -40.0 << -30.0 << -20.0 << -10.0 << 0.0 << 10.0 << 20.0 << 30.0 << 40.0 << 50.0;
        else
            leftValues << 5.0 << 10.0 << 15.0 << 20.0 << 25.0 << 30.0 << 35.0 << 40.0 << 45.0 << 50.0 << 55.0;
        unit = " " + tr("dB", "unit for decibel");
        break;
    case champ_initialFilterFc:
        if (isPrst)
        {
            leftValues << 0.01 << 0.1 << 1.0 << 10.0 << 100.0 << 1000.0 << 10000.0 << 100000.0;
            unit = "×";
        }
        else
        {
            leftValues << 20.0 << 50.0 << 100.0 << 200.0 << 500.0 << 1000.0 << 2000.0 << 5000.0 << 10000.0;
            unit = " " + tr("Hz", "unit for Herz");
        }
        break;
    case champ_initialFilterQ:
        if (isPrst)
            leftValues << -80.0 << -60.0 << -40.0 << -20.0 << 0.0 << 20.0 << 40.0 << 60.0 << 80.0;
        else
            leftValues << 10.0 << 20.0 << 30.0 << 40.0 << 50.0 << 60.0 << 70.0 << 80.0 << 90.0;
        unit = " " + tr("dB", "unit for decibel");
        break;
    case champ_attackVolEnv:
        if (isPrst)
        {
            leftValues << 0.01 << 0.1 << 1.0 << 10.0 << 100.0 << 1000.0 << 10000.0 << 100000.0;
            unit = "×";
        }
        else
        {
            leftValues << 0.003 << 0.01 << 0.03 << 0.1 << 0.3 << 1.0 << 3.0 << 10.0 << 30.0 << 100.0;
            unit = " " + tr("s", "unit for second");
        }
        break;
    case champ_fineTune:
        if (isPrst)
            leftValues << -200.0 << -150.0 << -100.0 << -50.0 << 0.0 << 50.0 << 100.0 << 150.0 << 200.0;
        else
            leftValues << -100.0 << -75.0 << -50.0 << -25.0 << 0.0 << 25.0 << 50.0 << 75.0 << 100.0;
        unit = " " + tr("cents");
        break;
    default:
        break;
    }

    // Add horizontal lines and labels
    for (int i = 0; i < leftValues.count(); i++)
    {
        // Real value
        double rValue = leftValues[i];

        // Text
        _leftLabels[i]->setText(QString::number(rValue) + unit);

        // Position
        double nValue = ExtensionView_Bars_Graphics_RectangleItem::getNormalizedValue(rValue, attributeType, isPrst);
        _leftLabels[i]->setY(1.0 - nValue);
        _hLines[i]->setY(1.0 - nValue);
    }
}

void ExtensionView_Bars_Graphics::showEvent(QShowEvent * event)
{
    QGraphicsView::showEvent(event);
    updateLabelPosition();
}

void ExtensionView_Bars_Graphics::resizeEvent(QResizeEvent * event)
{
    QGraphicsView::resizeEvent(event);
    fitInView(_displayedRect);
}

void ExtensionView_Bars_Graphics::mousePressEvent(QMouseEvent *event)
{
    // Alt key may have not been caught yet
    if (event->modifiers() == Qt::AltModifier && _currentMode == NORMAL_MODE)
        _currentMode = CONTOUR_MODE;

    switch (_currentMode)
    {
    case NORMAL_MODE:
        switch (event->button())
        {
        case Qt::LeftButton: {
            // Remember the initial y
            _yInit = this->mapToScene(event->pos()).y();

            // Rectangles below the mouse?
            QList<QList<ExtensionView_Bars_Graphics_RectangleItem*> > pairs = getRectanglesUnderMouse(event->pos());

            if (pairs.empty())
            {
                if (QApplication::keyboardModifiers() != Qt::ControlModifier && QApplication::keyboardModifiers() != Qt::ShiftModifier)
                {
                    // Remove the selection
                    bool reselect = false;
                    foreach (ExtensionView_Bars_Graphics_RectangleItem * item, _rectangles)
                    {
                        if (item->isSelected())
                        {
                            item->setSelected(false);
                            reselect = true;
                        }
                    }

                    // Update the selection outside the range editor
                    if (reselect)
                        triggerDivisionSelected();
                }
            }
            else
            {
                this->setCursor(Qt::ClosedHandCursor);
                _currentMode = EDITING_MODE;
                ExtensionView_Bars_Graphics_RectangleItem::syncHover(true);

                // Get the highlighted pair, or the first one
                int currentPairIndex = -1;
                for (int i = 0; i < pairs.count(); i++)
                {
                    foreach (ExtensionView_Bars_Graphics_RectangleItem * item, pairs[i])
                    {
                        if (item->isHovered())
                        {
                            currentPairIndex = i;
                            break;
                        }
                    }
                    if (currentPairIndex != -1)
                        break;
                }
                if (currentPairIndex == -1)
                    currentPairIndex = 0;

                // Is the current pair selected?
                bool currentIsSelected = false;
                bool reselect = false;
                foreach (ExtensionView_Bars_Graphics_RectangleItem * item, pairs[currentPairIndex])
                {
                    if (item->isSelected())
                    {
                        currentIsSelected = true;

                        // Possibly complete the selection
                        if (ContextManager::configuration()->getValue(ConfManager::SECTION_NONE, "stereo_modification", true).toBool())
                        {
                            // Select the whole pair
                            foreach (ExtensionView_Bars_Graphics_RectangleItem * item, pairs[currentPairIndex])
                            {
                                if (!item->isSelected())
                                {
                                    item->setSelected(true);
                                    reselect = true;
                                }
                            }
                        }
                        break;
                    }
                }
                if (reselect)
                    triggerDivisionSelected();

                Qt::KeyboardModifiers modifiers = QApplication::keyboardModifiers();
                if (modifiers == Qt::ShiftModifier)
                {
                    _removeIndexOnRelease = false;
                    _removeOtherIndexOnRelease = false;

                    // Select all rectangles between the base rectangle and the new one
                    if (_firstShiftRect.x() > -1)
                    {
                        QRectF shiftZone = _firstShiftRect.united(pairs[currentPairIndex][0]->rect());
                        foreach (ExtensionView_Bars_Graphics_RectangleItem * item, _rectangles)
                            item->setSelected(shiftZone.intersects(item->rect()));
                    }
                    else if (!currentIsSelected)
                    {
                        // Select the current pair
                        foreach (ExtensionView_Bars_Graphics_RectangleItem * item, pairs[currentPairIndex])
                            item->setSelected(true);
                    }

                    // Update the selection outside the range editor
                    triggerDivisionSelected();
                }
                else if (modifiers == Qt::ControlModifier)
                {
                    _removeIndexOnRelease = currentIsSelected;
                    _removeOtherIndexOnRelease = false;

                    if (!currentIsSelected)
                    {
                        // Select the current pair
                        foreach (ExtensionView_Bars_Graphics_RectangleItem * item, pairs[currentPairIndex])
                        {
                            item->setSelected(true);
                            _firstShiftRect = item->rect();
                        }

                        // Update the selection outside the range editor
                        triggerDivisionSelected();
                    }
                }
                else // No modifiers
                {
                    _removeIndexOnRelease = false;
                    _removeOtherIndexOnRelease = true;

                    if (!currentIsSelected)
                    {
                        // Deselect everything
                        foreach (ExtensionView_Bars_Graphics_RectangleItem * item, _rectangles)
                            item->setSelected(false);

                        // Select the current pair
                        foreach (ExtensionView_Bars_Graphics_RectangleItem * item, pairs[currentPairIndex])
                        {
                            item->setSelected(true);
                            _firstShiftRect = item->rect();
                        }

                        // Update the selection outside the range editor
                        triggerDivisionSelected();
                    }
                }
            }
            _currentMode = EDITING_MODE;
        } break;
        case Qt::MiddleButton: {
            // Trigger a key
            QPointF p = this->mapToScene(event->pos());
            int key = qRound(p.x());
            QApplication::postEvent(ContextManager::midi(), new NoteEvent(-1, key, 127));
            _keyTriggered = key;
        } break;
        default:
            break;
        }
        break;
    case EDITING_MODE:
        // Another mouse key is triggered, back to the normal mode
        _currentMode = NORMAL_MODE;
        break;
    case EDITING_MODE_WITH_MOVE:
        // Cancel the current editing
        foreach (ExtensionView_Bars_Graphics_RectangleItem * item, _rectangles)
            if (item->isSelected())
                item->cancelChanges();
        _legendItem2->setNewValue(false, 0.0);

        // Back to the normal mode
        _currentMode = NORMAL_MODE;
        break;
    case CONTOUR_MODE: {
        switch (event->button())
        {
        case Qt::LeftButton:
            this->setCursor(Qt::PointingHandCursor);
            setHeightForRectanglesBelowMouse(event->pos());
            _currentMode = EDITING_CONTOUR_MODE;
            break;
        default:
            break;
        }
    } break;
    case EDITING_CONTOUR_MODE:
        // Cancel the current editing
        foreach (ExtensionView_Bars_Graphics_RectangleItem * item, _rectangles)
            item->cancelChanges();
        _currentMode = CONTOUR_MODE;
        break;
    case PLAY_MODE:
        // Nothing
        break;
    }

    updateHover(event->pos());
    viewport()->update();
    QGraphicsView::mousePressEvent(event);
}

void ExtensionView_Bars_Graphics::mouseReleaseEvent(QMouseEvent *event)
{
    switch (_currentMode)
    {
    case NORMAL_MODE:
        switch (event->button())
        {
        case Qt::RightButton: {
            // Possibly change the hover
            QList<QList<ExtensionView_Bars_Graphics_RectangleItem*> > pairs = getRectanglesUnderMouse(event->pos());
            if (!pairs.empty())
            {
                // Last hovered index
                int lastHoverIndex = 0;
                for (int i = 0; i < pairs.count(); i++)
                {
                    for (int j = 0; j < pairs[i].count(); j++)
                    {
                        if (pairs[i][j]->isHovered())
                        {
                            lastHoverIndex = i;
                            break;
                        }
                    }
                }

                // The next pair has the hover
                lastHoverIndex = (lastHoverIndex + 1) % pairs.count();

                // Change the hovered rectangles
                foreach (ExtensionView_Bars_Graphics_RectangleItem * rectangle, _rectangles)
                    rectangle->setHover(pairs[lastHoverIndex].contains(rectangle));
            }
        } break;
        default:
            break;
        }
        break;
    case EDITING_MODE: {
        switch (event->button())
        {
        case Qt::LeftButton: {
            // Change the selection
            QList<QList<ExtensionView_Bars_Graphics_RectangleItem*> > pairs = getRectanglesUnderMouse(event->pos());
            this->setCursor(pairs.empty() ? Qt::ArrowCursor : Qt::OpenHandCursor);
            if (!pairs.empty())
            {
                // Current hovered and selected rectangles under the mouse
                int pairIndex = -1;
                for (int i = 0; i < pairs.count(); i++)
                {
                    for (int j = 0; j < pairs[i].count(); j++)
                    {
                        if (pairs[i][j]->isHovered() && pairs[i][j]->isSelected())
                        {
                            pairIndex = i;
                            break;
                        }
                    }
                    if (pairIndex != -1)
                        break;
                }

                bool reselect = false;

                // Possibly deselect them
                if (_removeIndexOnRelease)
                {
                    for (int i = 0; i < pairs[pairIndex].count(); i++)
                    {
                        if (pairs[pairIndex][i]->isSelected())
                        {
                            pairs[pairIndex][i]->setSelected(false);
                            reselect = true;
                        }
                    }

                    // Update the selection outside the range editor
                    triggerDivisionSelected();
                }

                // Possibly deselect all but them
                if (_removeOtherIndexOnRelease)
                {
                    // Deselect all but the current hovered rectangles
                    for (int i = 0; i < _rectangles.count(); i++)
                    {
                        if (!pairs[pairIndex].contains(_rectangles[i]))
                        {
                            if (_rectangles[i]->isSelected())
                            {
                                _rectangles[i]->setSelected(false);
                                reselect = true;
                            }
                        }
                    }

                    // Update the selection outside the range editor
                    if (reselect)
                        triggerDivisionSelected();
                }
            }
            _currentMode = NORMAL_MODE;
        } break;
        default:
            break;
        }
    } break;
    case EDITING_MODE_WITH_MOVE: {
        switch (event->button())
        {
        case Qt::LeftButton: {
            QList<QList<ExtensionView_Bars_Graphics_RectangleItem*> > pairs = getRectanglesUnderMouse(event->pos());
            this->setCursor(pairs.empty() ? Qt::ArrowCursor : Qt::OpenHandCursor);

            // Save the changes
            bool withChanges = false;
            foreach (ExtensionView_Bars_Graphics_RectangleItem * item, _rectangles)
                if (item->isSelected())
                    withChanges |= item->saveChanges();

            if (withChanges)
                _sf2->endEditing("barEditor");

            _legendItem2->setNewValue(false, 0.0);
            _currentMode = NORMAL_MODE;
        } break;
        default:
            break;
        }
    } break;
    case CONTOUR_MODE:
       break;
    case EDITING_CONTOUR_MODE:
        switch (event->button())
        {
        case Qt::LeftButton: {
            // Save the changes
            bool withChanges = false;
            foreach (ExtensionView_Bars_Graphics_RectangleItem * item, _rectangles)
                withChanges |= item->saveChanges();
            if (withChanges)
                _sf2->endEditing("barEditor");
            _currentMode = CONTOUR_MODE;
            QList<QList<ExtensionView_Bars_Graphics_RectangleItem*> > pairs = getRectanglesUnderMouse(event->pos());
            this->setCursor(pairs.empty() ? Qt::ArrowCursor : Qt::OpenHandCursor);
        } break;
        default:
            break;
        }
        break;
    case PLAY_MODE: {
        switch (event->button())
        {
        case Qt::MiddleButton:
            // Stop a key
            if (_keyTriggered != -1)
            {
                QApplication::postEvent(ContextManager::midi(), new NoteEvent(-1, _keyTriggered, 0));
                _keyTriggered = -1;
            }
            _currentMode = NORMAL_MODE;
            break;
        default:
            break;
        }
    } break;
    }

    ExtensionView_Bars_Graphics_RectangleItem::syncHover(false);
    updateHover(event->pos());
    viewport()->update();
    QGraphicsView::mouseReleaseEvent(event);
}

void ExtensionView_Bars_Graphics::mouseMoveEvent(QMouseEvent *event)
{
    switch (_currentMode)
    {
    case NORMAL_MODE: {
        QList<QList<ExtensionView_Bars_Graphics_RectangleItem*> > pairs = getRectanglesUnderMouse(event->pos());
        this->setCursor(pairs.empty() ? Qt::ArrowCursor : Qt::OpenHandCursor);

        // Section of the legend
        bool isLeft = (event->pos().x() > this->width() / 2);
        if (_legendItem->isLeft() != isLeft)
        {
            _legendItem->setLeft(isLeft);
            _legendItem2->setLeft(isLeft);
            updateLabelPosition();
        }

        // Update legend content
        updateHover(event->pos());
    } break;
    case EDITING_MODE: case EDITING_MODE_WITH_MOVE: {
        // Try to move rectangles
        ExtensionView_Bars_Graphics_RectangleItem * highlightedRectangle = nullptr;
        foreach (ExtensionView_Bars_Graphics_RectangleItem * item, _rectangles)
        {
            if (item->isSelected())
            {
                item->setNewValue(_yInit - this->mapToScene(event->pos()).y());
                if (item->isHovered())
                    highlightedRectangle = item;
            }
        }

        if (highlightedRectangle != nullptr)
        {
            // Update the modification legend
            _legendItem2->setNewValue(true, highlightedRectangle->currentVal());
        }

        _currentMode = EDITING_MODE_WITH_MOVE;
    } break;
    case EDITING_CONTOUR_MODE:
        setHeightForRectanglesBelowMouse(event->pos());
        break;
    case PLAY_MODE: case CONTOUR_MODE:
        break;
    }

    viewport()->update();
    QGraphicsView::mouseMoveEvent(event);
}

void ExtensionView_Bars_Graphics::mouseDoubleClickEvent(QMouseEvent * event)
{
    if (event->button() == Qt::LeftButton && _currentMode == NORMAL_MODE)
    {
        // Search the focussed divisions
        QList<QList<ExtensionView_Bars_Graphics_RectangleItem*> > pairs = getRectanglesUnderMouse(event->pos());
        IdList divIds;
        for (int i = 0; i < pairs.count(); i++)
        {
            for (int j = 0; j < pairs[i].count(); j++)
            {
                if (pairs[i][j]->isHovered())
                {
                    for (int k = 0; k < pairs[i].count(); k++)
                        divIds << pairs[i][k]->getID();
                    break;
                }
                if (!divIds.empty())
                    break;
            }
        }

        if (divIds.empty() && !pairs.isEmpty())
        {
            for (int i = 0; i < pairs[0].count(); i++)
                divIds << pairs[0][i]->getID();
        }

        // Get the child ids
        IdList childIds;
        foreach (EltID divId, divIds)
        {
            EltID childId = divId;
            if (divId.typeElement == elementInstSmpl)
            {
                childId.typeElement = elementSmpl;
                childId.indexElt = _sf2->get(divId, champ_sampleID).wValue;
            }
            else if (divId.typeElement == elementPrstInst)
            {
                childId.typeElement = elementInst;
                childId.indexElt = _sf2->get(divId, champ_instrument).wValue;
            }
            else
                continue;
            childIds << childId;
        }

        // Change the selection
        if (!childIds.empty())
            emit(selectionChanged(childIds));
    }
}

void ExtensionView_Bars_Graphics::keyPressEvent(QKeyEvent *event)
{
    switch (_currentMode)
    {
    case NORMAL_MODE: case CONTOUR_MODE:
        if (_currentMode == CONTOUR_MODE && event->key() != Qt::Key_Down && event->key() != Qt::Key_Up)
            return;

        switch (event->key())
        {
        case Qt::Key_Down: case Qt::Key_Up: {
            ExtensionView_Bars_Graphics_RectangleItem::syncHover(true);
            ExtensionView_Bars_Graphics_RectangleItem::StepType stepType = ExtensionView_Bars_Graphics_RectangleItem::MIDDLE_STEP;
            if (event->modifiers() == Qt::ShiftModifier)
                stepType = ExtensionView_Bars_Graphics_RectangleItem::BIG_STEP;
            else if (event->modifiers() == Qt::AltModifier)
                stepType = ExtensionView_Bars_Graphics_RectangleItem::SMALL_STEP;

            ExtensionView_Bars_Graphics_RectangleItem * lastSelectedRectangle = nullptr;
            ExtensionView_Bars_Graphics_RectangleItem * lastSelectedRectangleWithHover = nullptr;
            foreach (ExtensionView_Bars_Graphics_RectangleItem * rectangle, _rectangles)
            {
                if (rectangle->isSelected())
                {
                    rectangle->addStep(stepType, event->key() == Qt::Key_Up);
                    lastSelectedRectangle = rectangle;

                    if (rectangle->isHovered())
                        lastSelectedRectangleWithHover = rectangle;
                }
            }
            if (lastSelectedRectangleWithHover != nullptr)
                lastSelectedRectangle = lastSelectedRectangleWithHover;

            if (lastSelectedRectangle != nullptr)
            {
                _legendItem2->setNewValue(true, lastSelectedRectangle->currentVal());
                this->updateHover(this->mapFromScene(lastSelectedRectangle->rect().center()));
                viewport()->update();
            }
        }
            break;
        case Qt::Key_Space: {
            // Select or unselect the hovered rectangles
            bool selectionChanged = false;
            foreach (ExtensionView_Bars_Graphics_RectangleItem * rectangle, _rectangles)
            {
                if (rectangle->isHovered())
                {
                    rectangle->setSelected(!rectangle->isSelected());
                    selectionChanged = true;
                }
            }
            if (selectionChanged)
                triggerDivisionSelected();
        } break;
        case Qt::Key_Right: case Qt::Key_Left: {
            QList<ExtensionView_Bars_Graphics_RectangleItem *> rectanglesToSelect = getNextUnselectedRectangles(event->key() == Qt::Key_Right);
            if (!rectanglesToSelect.isEmpty())
            {
                // Possibly unselect everything
                if (event->modifiers() != Qt::ShiftModifier)
                {
                    foreach (ExtensionView_Bars_Graphics_RectangleItem * rectangle, _rectangles)
                        rectangle->setSelected(false);
                }

                // Select the new rectangles
                ExtensionView_Bars_Graphics_RectangleItem * lastSelectedRectangle = nullptr;
                foreach (ExtensionView_Bars_Graphics_RectangleItem * rectangle, rectanglesToSelect)
                {
                    rectangle->setSelected(true);
                    lastSelectedRectangle = rectangle;
                }

                // Notify the selection changed
                triggerDivisionSelected();

                if (lastSelectedRectangle != nullptr)
                {
                    this->updateHover(this->mapFromScene(lastSelectedRectangle->rect().center()));
                    viewport()->update();
                }
            }
        } break;
        case Qt::Key_0: case Qt::Key_1: case Qt::Key_2: case Qt::Key_3: case Qt::Key_4:
        case Qt::Key_5: case Qt::Key_6: case Qt::Key_7: case Qt::Key_8: case Qt::Key_9: {
            float percent = event->key() == Qt::Key_0 ? 1.0f : static_cast<float>(event->key() - Qt::Key_1) * 0.1f;
            foreach (ExtensionView_Bars_Graphics_RectangleItem * rectangle, _rectangles)
                rectangle->setPercent(percent);

            // Save the changes
            bool withChanges = false;
            foreach (ExtensionView_Bars_Graphics_RectangleItem * item, _rectangles)
                withChanges |= item->saveChanges();
            if (withChanges)
                _sf2->endEditing("barEditor");

            viewport()->update();
        } break;
        case Qt::Key_Alt:
            // Contour mode
            _currentMode = CONTOUR_MODE;
            break;
        default:
            break;
        }
        break;
    case EDITING_MODE:
    case EDITING_MODE_WITH_MOVE:
    case EDITING_CONTOUR_MODE:
    case PLAY_MODE:
        break;
    }

    QGraphicsView::keyPressEvent(event);
}

void ExtensionView_Bars_Graphics::keyReleaseEvent(QKeyEvent *event)
{
    switch (_currentMode)
    {
    case NORMAL_MODE: case CONTOUR_MODE:
        if ((event->key() == Qt::Key_Down || event->key() == Qt::Key_Up) && !event->isAutoRepeat())
        {
            ExtensionView_Bars_Graphics_RectangleItem::syncHover(false);
            bool withChanges = false;
            ExtensionView_Bars_Graphics_RectangleItem * lastSelectedRectangle = nullptr;
            ExtensionView_Bars_Graphics_RectangleItem * lastSelectedRectangleWithHover = nullptr;
            foreach (ExtensionView_Bars_Graphics_RectangleItem * rectangle, _rectangles)
            {
                if (rectangle->isSelected())
                {
                    withChanges |= rectangle->saveChanges();
                    lastSelectedRectangle = rectangle;
                    if (rectangle->isHovered())
                        lastSelectedRectangleWithHover = rectangle;
                }
            }
            if (lastSelectedRectangleWithHover != nullptr)
                lastSelectedRectangle = lastSelectedRectangleWithHover;

            if (withChanges)
            {
                _sf2->endEditing("barEditor");

                _legendItem2->setNewValue(false, 0.0);
                this->updateHover(this->mapFromScene(lastSelectedRectangle->rect().center()));
                viewport()->update();
            }
        }
        else if (event->key() == Qt::Key_Alt)
        {
            _currentMode = NORMAL_MODE;
        }
        break;
    case EDITING_CONTOUR_MODE:
        if (event->key() == Qt::Key_Alt)
        {
            QPoint mousePos = QWidget::mapFromGlobal(QCursor::pos());
            QList<QList<ExtensionView_Bars_Graphics_RectangleItem*> > pairs = getRectanglesUnderMouse(mousePos);
            this->setCursor(pairs.empty() ? Qt::ArrowCursor : Qt::OpenHandCursor);
            _currentMode = NORMAL_MODE;
            updateHover(mousePos);
            viewport()->update();
        }
        break;
    case EDITING_MODE: case EDITING_MODE_WITH_MOVE: case PLAY_MODE:
        break;
    }

    QGraphicsView::keyReleaseEvent(event);
}

void ExtensionView_Bars_Graphics::wheelEvent(QWheelEvent * event)
{
    int delta = event->phase() == Qt::ScrollUpdate ? -event->pixelDelta().y() : event->angleDelta().y();
    if ((_currentMode == NORMAL_MODE || _currentMode == CONTOUR_MODE) && delta != 0)
    {
        ExtensionView_Bars_Graphics_RectangleItem::StepType stepType = ExtensionView_Bars_Graphics_RectangleItem::MIDDLE_STEP;
        if (event->modifiers() == Qt::ShiftModifier)
            stepType = ExtensionView_Bars_Graphics_RectangleItem::BIG_STEP;
        else if (event->modifiers() == Qt::AltModifier)
            stepType = ExtensionView_Bars_Graphics_RectangleItem::SMALL_STEP;

        bool withChanges = false;
        foreach (ExtensionView_Bars_Graphics_RectangleItem * rectangle, _rectangles)
        {
            if (rectangle->isHovered())
            {
                rectangle->addStep(stepType, delta >= 0);
                withChanges |= rectangle->saveChanges();
            }
        }

        if (withChanges)
        {
            _sf2->endEditing("barEditor");
            this->updateHover(event->position().toPoint());
            viewport()->update();
        }
    }

    QGraphicsView::wheelEvent(event);
}

QList<QList<ExtensionView_Bars_Graphics_RectangleItem*> > ExtensionView_Bars_Graphics::getRectanglesUnderMouse(QPoint mousePos)
{
    // Rectangles under the mouse
    QList<ExtensionView_Bars_Graphics_RectangleItem *> rectanglesUnderMouse;
    foreach (ExtensionView_Bars_Graphics_RectangleItem * item, _rectangles)
        if (item->contains(this->mapToScene(mousePos)))
            rectanglesUnderMouse << item;

    // Sort them by pairs
    QList<QList<ExtensionView_Bars_Graphics_RectangleItem *> > pairs;
    while (!rectanglesUnderMouse.isEmpty())
    {
        QList<ExtensionView_Bars_Graphics_RectangleItem *> listTmp;
        listTmp << rectanglesUnderMouse.takeFirst();
        EltID idBrother = ExtensionView_Bars_Graphics::findBrother(listTmp[0]->getID());
        foreach (ExtensionView_Bars_Graphics_RectangleItem * item, rectanglesUnderMouse)
        {
            if (*item == idBrother)
            {
                listTmp << item;
                rectanglesUnderMouse.removeOne(item);
                break;
            }
        }
        pairs << listTmp;
    }

    return pairs;
}

void ExtensionView_Bars_Graphics::playKey(int key, int velocity)
{
    if (velocity == 0 && _mapGraphicsKeys[key] != nullptr)
    {
        // A key is removed
        delete _mapGraphicsKeys.take(key);
    }
    else if (velocity > 0 && _mapGraphicsKeys[key] == nullptr)
    {
        // A key is added
        _mapGraphicsKeys[key] = new ExtensionView_Bars_Graphics_Key();
        _scene->addItem(_mapGraphicsKeys[key]);
        _mapGraphicsKeys[key]->setKey(key);
        _mapGraphicsKeys[key]->setZValue(25);
    }
}

void ExtensionView_Bars_Graphics::triggerDivisionSelected()
{
    IdList ids;
    foreach (ExtensionView_Bars_Graphics_RectangleItem * item, _rectangles)
        if (item->isSelected())
            ids << item->getID();
    if (ids.empty())
        ids << _defaultID;
    _keepFirstShiftPoint = true;
    emit(selectionChanged(ids));
    _keepFirstShiftPoint = false;
}

EltID ExtensionView_Bars_Graphics::findBrother(EltID id)
{
    if (ContextManager::configuration()->getValue(ConfManager::SECTION_NONE, "stereo_modification", true).toBool() &&
            id.typeElement == elementInstSmpl)
    {
        // Sample linked to the division
        SoundfontManager * sm = SoundfontManager::getInstance();
        EltID idSmpl = id;
        idSmpl.typeElement = elementSmpl;
        idSmpl.indexElt = sm->get(id, champ_sampleID).wValue;

        // Sample linked to another one?
        SFSampleLink typeLink = sm->get(idSmpl, champ_sfSampleType).sfLinkValue;
        if (typeLink == rightSample || typeLink == leftSample || typeLink == linkedSample ||
                typeLink == RomRightSample || typeLink == RomLeftSample || typeLink == RomLinkedSample)
        {
            // Characteristics to find in the other divisions
            int numSmpl2 = sm->get(idSmpl, champ_wSampleLink).wValue;
            RangesType keyRange = sm->get(id, champ_keyRange).rValue;
            RangesType velRange = sm->get(id, champ_velRange).rValue;

            // Search the brother
            int numBrothers = 0;
            EltID idBrother = id;
            EltID id2 = id;
            foreach (int i, sm->getSiblings(id))
            {
                id2.indexElt2 = i;
                if (i != id.indexElt2)
                {
                    RangesType keyRange2 = sm->get(id2, champ_keyRange).rValue;
                    RangesType velRange2 = sm->get(id2, champ_velRange).rValue;
                    if (keyRange2.byLo == keyRange.byLo && keyRange2.byHi == keyRange.byHi &&
                            velRange2.byLo == velRange.byLo && velRange2.byHi == velRange.byHi)
                    {
                        int iTmp = sm->get(id2, champ_sampleID).wValue;
                        if (iTmp == idSmpl.indexElt)
                            numBrothers = 2; // ambiguity, another sample is like id
                        else if (iTmp == numSmpl2)
                        {
                            numBrothers++;
                            idBrother.indexElt2 = i;
                        }
                    }
                }
            }

            // Return the brother if there is only one (no ambiguity)
            if (numBrothers == 1)
                return idBrother;
            else
                return EltID(elementUnknown, 0, 0, 0, 0);
        }
        else
            return EltID(elementUnknown, 0, 0, 0, 0);
    }
    else
        return EltID(elementUnknown, 0, 0, 0, 0);
}

QList<ExtensionView_Bars_Graphics_RectangleItem *> ExtensionView_Bars_Graphics::getNextUnselectedRectangles(bool toTheRight)
{
    // Find the rightmost or leftmost selected rectangle
    ExtensionView_Bars_Graphics_RectangleItem * extremeRect = nullptr;
    if (toTheRight)
    {
        foreach (ExtensionView_Bars_Graphics_RectangleItem * rectangle, _rectangles)
        {
            if (rectangle->isSelected())
            {
                if (extremeRect == nullptr || rectangle->boundingRect().right() > extremeRect->boundingRect().right())
                    extremeRect = rectangle;
            }
        }
    }
    else
    {
        foreach (ExtensionView_Bars_Graphics_RectangleItem * rectangle, _rectangles)
        {
            if (rectangle->isSelected())
            {
                if (extremeRect == nullptr || rectangle->boundingRect().left() < extremeRect->boundingRect().left())
                    extremeRect = rectangle;
            }
        }
    }

    QList<ExtensionView_Bars_Graphics_RectangleItem *> result;
    if (extremeRect == nullptr)
    {
        // Get the first or the last rectangle
        if (toTheRight)
        {
            foreach (ExtensionView_Bars_Graphics_RectangleItem * rectangle, _rectangles)
                if (extremeRect == nullptr || rectangle->boundingRect().left() < extremeRect->boundingRect().left())
                    extremeRect = rectangle;
            if (extremeRect != nullptr)
                result << extremeRect;
        }
        else
        {
            foreach (ExtensionView_Bars_Graphics_RectangleItem * rectangle, _rectangles)
                if (extremeRect == nullptr || rectangle->boundingRect().right() > extremeRect->boundingRect().right())
                    extremeRect = rectangle;
            if (extremeRect != nullptr)
                result << extremeRect;
        }
    }
    else
    {
        // Get the next unselected rectangle, if there is one
        ExtensionView_Bars_Graphics_RectangleItem * nextRectangle = nullptr;
        if (toTheRight)
        {
            double limit = extremeRect->rect().right();
            foreach (ExtensionView_Bars_Graphics_RectangleItem * rectangle, _rectangles)
                if (rectangle->rect().left() >= limit && (nextRectangle == nullptr || rectangle->rect().left() < nextRectangle->rect().left()))
                    nextRectangle = rectangle;
        }
        else
        {
            double limit = extremeRect->rect().left();
            foreach (ExtensionView_Bars_Graphics_RectangleItem * rectangle, _rectangles)
                if (rectangle->rect().right() <= limit && (nextRectangle == nullptr || rectangle->rect().right() > nextRectangle->rect().right()))
                    nextRectangle = rectangle;
        }
        if (nextRectangle != nullptr)
            result << nextRectangle;
    }

    // Stereo division?
    if (!result.isEmpty())
    {
        EltID brotherId = findBrother(result[0]->getID());
        if (brotherId.typeElement != elementUnknown)
        {
            foreach (ExtensionView_Bars_Graphics_RectangleItem * rectangle, _rectangles)
            {
                if (rectangle->getID() == brotherId)
                {
                    result << rectangle;
                    break;
                }
            }
        }
    }

    return result;
}

void ExtensionView_Bars_Graphics::setHeightForRectanglesBelowMouse(QPoint mousePos)
{
    QPointF scenePos = this->mapToScene(mousePos);
    foreach (ExtensionView_Bars_Graphics_RectangleItem * item, _rectangles)
    {
        if (item->contains(this->mapToScene(mousePos)))
            item->setPercent(1.0f - scenePos.y());
    }
}
