/***************************************************************************
**                                                                        **
**  Polyphone, a soundfont editor                                         **
**  Copyright (C) 2013-2020 Davy Triponney                                **
**                                                                        **
**  This program is free software: you can redistribute it and/or modify  **
**  it under the terms of the GNU General Public License as published by  **
**  the Free Software Foundation, either version 3 of the License, or     **
**  (at your option) any later version.                                   **
**                                                                        **
**  This program is distributed in the hope that it will be useful,       **
**  but WITHOUT ANY WARRANTY; without even the implied warranty of        **
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the          **
**  GNU General Public License for more details.                          **
**                                                                        **
**  You should have received a copy of the GNU General Public License     **
**  along with this program. If not, see http://www.gnu.org/licenses/.    **
**                                                                        **
****************************************************************************
**           Author: Davy Triponney                                       **
**  Website/Contact: https://www.polyphone-soundfonts.com                 **
**             Date: 01.01.2013                                           **
***************************************************************************/

#include "extensionview_bars_graphics_legenditem2.h"
#include "extensionview_bars_graphics_rectangleitem.h"
#include "contextmanager.h"
#include <QPainter>

const int ExtensionView_Bars_Graphics_LegendItem2::s_border = 5;

ExtensionView_Bars_Graphics_LegendItem2::ExtensionView_Bars_Graphics_LegendItem2(QString fontFamily, QGraphicsItem *parent) : QGraphicsItem(parent),
    _font(fontFamily, 8),
    _offsetY(0)
{
    // Colors
    QColor color = ContextManager::theme()->getFixedColor(ThemeManager::RED, ThemeManager::LIST_BACKGROUND);
    _borderPen = QPen(color, 1, Qt::SolidLine);
    color.setAlpha(200);
    _foregroundBrush = QBrush(color);
    _textPen = QPen(ContextManager::theme()->getColor(ThemeManager::HIGHLIGHTED_TEXT), 1, Qt::SolidLine);

    this->setFlag(QGraphicsItem::ItemIgnoresTransformations);
    setLeft(true);
}

void ExtensionView_Bars_Graphics_LegendItem2::setLeft(bool isLeft)
{
    if (isLeft)
        _alignment = Qt::AlignLeft | Qt::AlignTop;
    else
        _alignment = Qt::AlignRight | Qt::AlignTop;
}

void ExtensionView_Bars_Graphics_LegendItem2::setNewValue(bool isDisplayed, double value)
{
    _text.clear();
    if (isDisplayed)
    {
        // Real value
        double realValue = ExtensionView_Bars_Graphics_RectangleItem::getValue(value, _attributeType, _isPrst);
        AttributeValue value = Attribute(_attributeType, _isPrst, realValue).getStoredValue();

        QString textValue = QString::number(Attribute::toRealValue(_attributeType, _isPrst, value), 'f', 3);
        while (!textValue.isEmpty() && (textValue.endsWith("0")))
            textValue = textValue.left(textValue.size() - 1);
        if (!textValue.isEmpty() && (textValue.endsWith(".") || textValue.endsWith(",")))
            textValue = textValue.left(textValue.size() - 1);

        // Store the value
        _text << Attribute::getDescription(_attributeType, _isPrst) + ": " + textValue;
    }
}

QRectF ExtensionView_Bars_Graphics_LegendItem2::boundingRect() const
{
    QSizeF size = getTextSize();
    return QRectF(QPointF(dx(size), dy(size)), size);
}

void ExtensionView_Bars_Graphics_LegendItem2::paint(QPainter *painter, const QStyleOptionGraphicsItem * option, QWidget * widget)
{
    Q_UNUSED(option)
    Q_UNUSED(widget)
    if (_text.isEmpty())
        return;

    // Translation of the painter
    QSizeF size = getTextSize();
    painter->translate(dx(size), dy(size));

    // Paint foreground and border
    painter->setBrush(_foregroundBrush);
    painter->setPen(_borderPen);
    painter->drawRect(QRectF(QPoint(0, 0), size));

    // Paint text
    painter->setPen(_textPen);
    painter->setFont(_font);
    QFontMetrics fm(_font);
    for (int i = 0; i < _text.count(); i++)
        painter->drawText(QPoint(s_border, fm.height() * (i + 1)), _text.at(i));
}


QSizeF ExtensionView_Bars_Graphics_LegendItem2::getTextSize() const
{
    if (_text.isEmpty())
        return QSizeF(0, 0);

    QFontMetrics fm(_font);

    int height = _text.count() * fm.height() + 2 * s_border;
    int width = 0;
    foreach (QString line, _text)
        width = qMax(width, fm.horizontalAdvance(line));
    width += 2 * s_border;

    return QSizeF(width, height);
}

qreal ExtensionView_Bars_Graphics_LegendItem2::dx(QSizeF size) const
{
    if (_alignment & Qt::AlignHCenter)
        return -size.width() / 2;
    else if (_alignment & Qt::AlignRight)
        return -size.width();
    return 0;
}

qreal ExtensionView_Bars_Graphics_LegendItem2::dy(QSizeF size) const
{
    double value = _offsetY;

    if (_alignment & Qt::AlignVCenter)
        value -= size.height() / 2;
    else if (_alignment & Qt::AlignBottom)
        value -= size.height();

    return value;
}
