/***************************************************************************
**                                                                        **
**  Polyphone, a soundfont editor                                         **
**  Copyright (C) 2013-2020 Davy Triponney                                **
**                                                                        **
**  This program is free software: you can redistribute it and/or modify  **
**  it under the terms of the GNU General Public License as published by  **
**  the Free Software Foundation, either version 3 of the License, or     **
**  (at your option) any later version.                                   **
**                                                                        **
**  This program is distributed in the hope that it will be useful,       **
**  but WITHOUT ANY WARRANTY; without even the implied warranty of        **
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the          **
**  GNU General Public License for more details.                          **
**                                                                        **
**  You should have received a copy of the GNU General Public License     **
**  along with this program. If not, see http://www.gnu.org/licenses/.    **
**                                                                        **
****************************************************************************
**           Author: Davy Triponney                                       **
**  Website/Contact: https://www.polyphone-soundfonts.com                 **
**             Date: 01.01.2013                                           **
***************************************************************************/

#include "extensionview_bars_graphics_rectangleitem.h"
#include "contextmanager.h"
#include "soundfontmanager.h"
#include <QPainter>
#include "qmath.h"

bool ExtensionView_Bars_Graphics_RectangleItem::s_synchronizeEditingMode = false;
const double ExtensionView_Bars_Graphics_RectangleItem::BAR_THICKNESS = 0.6;

ExtensionView_Bars_Graphics_RectangleItem::ExtensionView_Bars_Graphics_RectangleItem(EltID id, AttributeType attributeType, QGraphicsItem *parent) : QGraphicsRectItem(parent),
    _id(id),
    _attributeType(attributeType),
    _editingMode(false),
    _isSelected(false)
{
    // Initialize pens and brushes
    QColor color = ContextManager::theme()->getColor(ThemeManager::HIGHLIGHTED_BACKGROUND);
    color.setAlpha(180);
    _penBorderThin = QPen(color, 1);
    color.setAlpha(255);
    _penBorderFat = QPen(color, 3);
    color.setAlpha(60);
    _brushRectangle = QBrush(color);
    color.setAlpha(140);
    _brushRectangleSelected = QBrush(color, Qt::DiagCrossPattern);
    _brushRectangleSelected.setTransform(QTransform(1,0,1,0,0,0)); // Don't scale the pattern

    initialize(_id, _attributeType);
    _penBorderThin.setCosmetic(true);
    _penBorderFat.setCosmetic(true);

    this->setRect(this->getRectF());
}

void ExtensionView_Bars_Graphics_RectangleItem::initialize(EltID id, AttributeType attributeType)
{
    // Default values
    _minKey = 0;
    _maxKey = 127;
    _currentVal = 0;

    // Global division
    EltID idGlobal = id;
    if (id.typeElement == elementInstSmpl)
        idGlobal.typeElement = elementInst;
    else
        idGlobal.typeElement = elementPrst;

    // Key range
    SoundfontManager * sm = SoundfontManager::getInstance();
    if (sm->isSet(id, champ_keyRange))
    {
        RangesType range = sm->get(id, champ_keyRange).rValue;
        _minKey = range.byLo;
        _maxKey = range.byHi;
    }
    else if (sm->isSet(idGlobal, champ_keyRange))
    {
        RangesType range = sm->get(idGlobal, champ_keyRange).rValue;
        _minKey = range.byLo;
        _maxKey = range.byHi;
    }

    // Attribute value
    AttributeValue value;
    if (sm->isSet(id, attributeType))
        value = sm->get(id, attributeType);
    else if (sm->isSet(idGlobal, attributeType))
        value = sm->get(idGlobal, attributeType);
    else
        value = Attribute::getDefaultStoredValue(attributeType, id.typeElement == elementPrst || id.typeElement == elementPrstInst);
    bool isPrst = id.typeElement == elementPrst || id.typeElement == elementPrstInst;
    double realValue = Attribute(attributeType, isPrst, value).getRealValue();

    // Normalize between 0 and 1
    _currentVal = getNormalizedValue(realValue, attributeType, isPrst);
    _currentValInit = _currentVal;
}

bool ExtensionView_Bars_Graphics_RectangleItem::contains(const QPointF &point) const
{
    // Only a check on x is made
    return -0.5 + _minKey <= point.x() && point.x() < 0.5 + _maxKey;
}

void ExtensionView_Bars_Graphics_RectangleItem::setHover(bool isHovered)
{
    if (isHovered)
    {
        _editingMode = true;
        this->setZValue(51);
    }
    else
    {
        _editingMode = false;
        this->setZValue(50);
    }
}

void ExtensionView_Bars_Graphics_RectangleItem::paint(QPainter * painter, const QStyleOptionGraphicsItem * option, QWidget * widget)
{
    // Draw base rectangle with a background color and a thin border
    this->setBrush(_isSelected ? _brushRectangleSelected : _brushRectangle);
    this->setPen(_penBorderThin);
    QGraphicsRectItem::paint(painter, option, widget);

    QRectF rectF = this->rect();
    if ((_isSelected && s_synchronizeEditingMode) || _editingMode)
    {
        painter->setPen(_penBorderFat);
        painter->drawLine(rectF.topRight(), rectF.bottomRight());
        painter->drawLine(rectF.topLeft(), rectF.bottomLeft());
        painter->drawLine(rectF.topLeft(), rectF.topRight());
        painter->drawLine(rectF.bottomLeft(), rectF.bottomRight());
    }
}

double ExtensionView_Bars_Graphics_RectangleItem::getNormalizedValue(double value, AttributeType attributeType, bool isPrst)
{
    // Normalize between 0 and 1
    double normalizedValue;
    switch (attributeType)
    {
    case champ_initialAttenuation:
        normalizedValue = isPrst ? (value / 116.0 + 0.5) : (value / 58.0);
        //normalizedValue = 1.0 - normalizedValue;
        break;
    case champ_attackVolEnv:
        normalizedValue = isPrst ? ((log10(value) + 3.0) / 8.3) : ((log10(value) + 3.0) / 5.1);
        break;
    case champ_initialFilterFc:
        normalizedValue = isPrst ? ((log10(value) + 3.0) / 8.3) : ((log10(value) - 1.0) / 3.3);
        break;
    case champ_initialFilterQ:
        normalizedValue = isPrst ? (value / 192.0 + 0.5) : (value / 96.0);
        break;
    case champ_fineTune:
        normalizedValue = isPrst ? (value / 396.0 + 0.5) : (value / 198.0 + 0.5);
        break;
    default:
        normalizedValue = 0;
        break;
    }

    return normalizedValue;
}

double ExtensionView_Bars_Graphics_RectangleItem::getValue(double normalizedValue, AttributeType attributeType, bool isPrst)
{
    double value = 0;
    switch (attributeType)
    {
    case champ_initialAttenuation:
        //normalizedValue = 1.0 - normalizedValue;
        value = isPrst ? ((normalizedValue - 0.5) * 116.0) : (normalizedValue * 58.0);
        break;
    case champ_attackVolEnv:
        value = isPrst ? pow(10.0, normalizedValue * 8.3 - 3.0) : pow(10.0, normalizedValue * 5.1 - 3.0);
        break;
    case champ_initialFilterFc:
        value = isPrst ? pow(10.0, normalizedValue * 8.3 - 3.0) : pow(10.0, normalizedValue * 3.3 + 1.0);
        break;
    case champ_initialFilterQ:
        value = isPrst ? ((normalizedValue - 0.5) * 192.0) : (normalizedValue * 96.0);
        break;
    case champ_fineTune:
        value = isPrst ? ((normalizedValue - 0.5) * 396.0) : ((normalizedValue - 0.5) * 198.0);
        break;
    default:
        value = 0;
        break;
    }

    return value;
}

void ExtensionView_Bars_Graphics_RectangleItem::setNewValue(double delta)
{
    _currentVal = _currentValInit + delta;
    if (_currentVal < 0)
        _currentVal = 0;
    else if (_currentVal > 1)
        _currentVal = 1;

    this->setRect(getRectF());
}

void ExtensionView_Bars_Graphics_RectangleItem::addStep(StepType stepType, bool isPositive)
{
    bool isPrst = _id.typeElement == elementPrst || _id.typeElement == elementPrstInst;

    // Delta
    double delta;
    switch (_attributeType)
    {
    case champ_initialAttenuation:
        delta = isPrst ? 0.5 / 116 : 0.5 / 58.0;
        if (stepType == SMALL_STEP)
            delta *= 0.2;
        else if (stepType == BIG_STEP)
            delta *= 4;
        break;
    case champ_attackVolEnv: case champ_initialFilterFc:
        delta = isPrst ? 0.005 : 0.01;
        if (stepType == SMALL_STEP)
            delta *= 0.2;
        else if (stepType == BIG_STEP)
            delta *= 5;
        break;
    case champ_initialFilterQ:
        delta = isPrst ? 0.5 / 192.0 : 0.5 / 96.0;
        if (stepType == SMALL_STEP)
            delta *= 0.2;
        else if (stepType == BIG_STEP)
            delta *= 4;
        break;
    case champ_fineTune:
        delta = isPrst ? 2.0 / 396 : 2.0 / 198.0;
        if (stepType == SMALL_STEP)
            delta *= 0.5;
        else if (stepType == BIG_STEP)
            delta *= 2.5;
        break;
    default:
        delta = 0;
        break;
    }

    if (!isPositive)
        delta *= -1;

    _currentVal += delta;
    if (_currentVal < 0)
        _currentVal = 0;
    else if (_currentVal > 1)
        _currentVal = 1;

    this->setRect(getRectF());
}

void ExtensionView_Bars_Graphics_RectangleItem::setPercent(float percent)
{
    _currentVal = percent;
    this->setRect(getRectF());
}

bool ExtensionView_Bars_Graphics_RectangleItem::saveChanges()
{
    // Store the attribute value
    SoundfontManager * sm = SoundfontManager::getInstance();
    bool hasChanged = false;
    if (_currentVal != _currentValInit)
    {
        hasChanged = true;

        // Real value
        bool isPrst = _id.typeElement == elementPrst || _id.typeElement == elementPrstInst;
        double realValue = this->getValue(_currentVal, _attributeType, isPrst);

        // Store the value
        AttributeValue value = Attribute(_attributeType, isPrst, realValue).getStoredValue();
        sm->set(_id, _attributeType, value);
    }

    _currentValInit = _currentVal;
    return hasChanged;
}

void ExtensionView_Bars_Graphics_RectangleItem::cancelChanges()
{
    _currentVal = _currentValInit;
    this->setRect(getRectF());
}

QRectF ExtensionView_Bars_Graphics_RectangleItem::getRectF() const
{
    return QRectF(-0.5 + _minKey + (1.0 - BAR_THICKNESS) / 2, 1.01 - _currentVal, 1.0 + _maxKey - _minKey + BAR_THICKNESS - 1.0, _currentVal + 0.01);
}
