/***************************************************************************
**                                                                        **
**  Polyphone, a soundfont editor                                         **
**  Copyright (C) 2013-2020 Davy Triponney                                **
**                                                                        **
**  This program is free software: you can redistribute it and/or modify  **
**  it under the terms of the GNU General Public License as published by  **
**  the Free Software Foundation, either version 3 of the License, or     **
**  (at your option) any later version.                                   **
**                                                                        **
**  This program is distributed in the hope that it will be useful,       **
**  but WITHOUT ANY WARRANTY; without even the implied warranty of        **
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the          **
**  GNU General Public License for more details.                          **
**                                                                        **
**  You should have received a copy of the GNU General Public License     **
**  along with this program. If not, see http://www.gnu.org/licenses/.    **
**                                                                        **
****************************************************************************
**           Author: Davy Triponney                                       **
**  Website/Contact: https://www.polyphone-soundfonts.com                 **
**             Date: 01.01.2013                                           **
***************************************************************************/

#include "extensionview_bars_page.h"
#include "ui_extensionview_bars_page.h"
#include "soundfontmanager.h"
#include "contextmanager.h"

#include "attribute.h"

ExtensionView_Bars_Page::ExtensionView_Bars_Page(QWidget *parent) : Page(parent, "barEditor"),
    ui(new Ui::ExtensionView_Bars_Page)
{
    ui->setupUi(this);
    connect(ui->graph, SIGNAL(selectionChanged(IdList)), this, SIGNAL(selectedIdsChanged(IdList)));

    // Info tooltip
    ui->labelInfo->setPixmap(ContextManager::theme()->getColoredSvg(":/icons/info.svg", QSize(24, 24), ThemeManager::WINDOW_TEXT));
}

ExtensionView_Bars_Page::~ExtensionView_Bars_Page()
{
    delete ui;
}

void ExtensionView_Bars_Page::updateInterface(QString editingSource)
{
    // Check if the new parents are the same
    IdList parentIds = _currentIds.getSelectedIds(elementInst);
    if (parentIds.empty())
        parentIds = _currentIds.getSelectedIds(elementPrst);
    bool sameElement = true;
    if (parentIds.count() == _currentParentIds.count())
    {
        for (int i = 0; i < parentIds.count(); i++)
        {
            if (parentIds[i] != _currentParentIds[i])
            {
                sameElement = false;
                break;
            }
        }
    }
    else
        sameElement = false;
    bool justSelection = (sameElement && editingSource == "command:selection");

    // Store the new parent ids
    _currentParentIds = parentIds;

    // Multiple instruments or presets?
    if (_currentParentIds.count() > 1)
        ui->stackedWidget->setCurrentIndex(1);
    else
    {
        // Update the graphics and the split button
        ui->graph->display(_currentIds, getAttribute(), justSelection);
        updatePushSplit(_currentIds);

        // Attribute names
        bool isPrst = (_currentIds[0].typeElement == elementPrst || _currentIds[0].typeElement == elementPrstInst);
        ui->radioAttenuation->setText(Attribute::getDescription(champ_initialAttenuation, isPrst));
        ui->radioAttack->setText(Attribute::getDescription(champ_attackVolEnv, isPrst));
        ui->radioFilterFrequency->setText(Attribute::getDescription(champ_initialFilterFc, isPrst));
        ui->radioFilterResonance->setText(Attribute::getDescription(champ_initialFilterQ, isPrst));

        ui->stackedWidget->setCurrentIndex(0);
    }
}

void ExtensionView_Bars_Page::keyPlayedInternal(int key, int velocity)
{
    ui->graph->playKey(key, velocity);
}

AttributeType ExtensionView_Bars_Page::getAttribute()
{
    if (ui->radioAttack->isChecked())
        return champ_attackVolEnv;
    if (ui->radioFilterFrequency->isChecked())
        return champ_initialFilterFc;
    if (ui->radioFilterResonance->isChecked())
        return champ_initialFilterQ;
    if (ui->radioTuning->isChecked())
        return champ_fineTune;
    return champ_initialAttenuation;
}

void ExtensionView_Bars_Page::onRadioButtonClicked()
{
    ui->graph->display(_currentIds, getAttribute(), false);
    ui->graph->setFocus();
}

void ExtensionView_Bars_Page::updatePushSplit(IdList ids)
{
    // Browse all ids and check if it is possible to split divisions
    bool isEnabled = false;
    foreach (EltID id, ids)
    {
        if (id.typeElement != elementInstSmpl && id.typeElement != elementPrstInst)
            continue;

        // Key range
        RangesType r = _sf2->get(id, champ_keyRange).rValue;
        if (r.byLo != r.byHi)
        {
            isEnabled = true;
            break;
        }
    }
    ui->pushSplit->setEnabled(isEnabled);
}

void ExtensionView_Bars_Page::on_pushSplit_clicked()
{
    // List of ids to split
    IdList ids;
    bool stereoEditing = ContextManager::configuration()->getValue(ConfManager::SECTION_NONE, "stereo_modification", true).toBool();
    foreach (EltID id, _currentIds)
    {
        if (id.typeElement != elementInstSmpl && id.typeElement != elementPrstInst)
            continue;

        // Key range
        RangesType r = _sf2->get(id, champ_keyRange).rValue;
        if (r.byLo != r.byHi)
        {
            if (!ids.contains(id))
                ids << id;

            // Stereo editing?
            if (id.typeElement == elementInstSmpl && stereoEditing)
            {
                EltID brotherId = ExtensionView_Bars_Graphics::findBrother(id);
                if (brotherId.typeElement == elementInstSmpl && !ids.contains(brotherId))
                    ids << brotherId;
            }
        }
    }
    if (ids.isEmpty())
        return;

    // Split all ids
    IdList idsToSelect;
    foreach (EltID id, ids)
        idsToSelect << splitId(id);
    _sf2->endEditing("barEditor:split");

    emit(selectedIdsChanged(idsToSelect));
}

IdList ExtensionView_Bars_Page::splitId(EltID id)
{
    IdList ids;
    ids << id;

    // Current key range
    RangesType r = _sf2->get(id, champ_keyRange).rValue;

    // Change keyrange
    AttributeValue val;
    val.rValue.byLo = r.byLo;
    val.rValue.byHi = r.byLo;
    _sf2->set(id, champ_keyRange, val);

    // Create identical divisions for other keys
    EltID id2 = id;
    for (int i = r.byLo + 1; i <= r.byHi; i++)
    {
        id2.indexElt2 = _sf2->add(id);
        ids << id2;

        // Copy gens and mods
        duplicateGenMod(id, id2);

        // Change keyrange
        val.rValue.byLo = i;
        val.rValue.byHi = i;
        _sf2->set(id2, champ_keyRange, val);
    }

    return ids;
}

void ExtensionView_Bars_Page::duplicateGenMod(EltID idFrom, EltID idTo)
{
    // Copy gens
    idFrom.typeElement = (idFrom.typeElement == elementInstSmpl ? elementInstSmplGen : elementPrstInstGen);
    foreach (int champ, _sf2->getSiblings(idFrom))
        _sf2->set(idTo, (AttributeType)champ, _sf2->get(idFrom.parent(), (AttributeType)champ));

    // Copy mods
    idFrom.typeElement = idTo.typeElement = (idFrom.typeElement == elementInstSmpl ? elementInstSmplMod : elementPrstInstMod);
    foreach (int j, _sf2->getSiblings(idFrom))
    {
        idFrom.indexMod = j;
        idTo.indexMod = _sf2->add(idTo);
        _sf2->set(idTo, champ_modAmount, _sf2->get(idFrom, champ_modAmount));
        _sf2->set(idTo, champ_sfModSrcOper, _sf2->get(idFrom, champ_sfModSrcOper));
        _sf2->set(idTo, champ_sfModAmtSrcOper, _sf2->get(idFrom, champ_sfModAmtSrcOper));
        _sf2->set(idTo, champ_sfModDestOper, _sf2->get(idFrom, champ_sfModDestOper));
        _sf2->set(idTo, champ_sfModTransOper, _sf2->get(idFrom, champ_sfModTransOper));
    }
}
